"""
Constants
==========================================

.. module:: nastro.constants

Physical and mathematical constants and units taken from SciPy.

======================================= ===============================================
``pi``, ``twopi``, ``halfpi``           Pi and multiples
``c``                                   speed of light in vacuum
``G``                                   Newtonian constant of gravitation
``degree``                              degree in radians
``arcmin``                              arc minute in radians
``arcsec``                              arc second in radians
``miliarcsec``                          milliarcsecond in radians
``minute``                              one minute in seconds
``hour``                                one hour in seconds
``day``                                 one day in seconds
``year``                                one year (365 days) in seconds
``julian_year``                         one Julian year (365.25 days) in seconds
``au``                                  one astronomical unit in meters
``light_year``                          one light year in meters
``parsec``                              one parsec in meters
``km``                                  kilometers in meters
``kmh``                                 kilometers per hour in meters per second
======================================= ===============================================
"""

import scipy.constants as sc

# Mathematical constants
pi = sc.pi
twopi = 2.0 * pi
halfpi = 0.5 * pi

# Physical constants
c = clight = sc.speed_of_light
G = sc.gravitational_constant

# Angles in radians
degree = pi / 180.0
arcmin = degree / 60.0
arcsec = arcmin / 60.0
miliarcsec = arcsec * 1e-3

# Time in seconds
minute = 60.0
hour = minute * 60.0
day = hour * 24.0
year = day * 365
julian_year = day * 365.25

# Lengths in meters
km = 1e3
au = astronomical_unit = sc.astronomical_unit
light_year = sc.light_year
parsec = sc.parsec
km = 1e3

# Speed in m/s
kmh = 1e3 / hour



from .. import types as nt


class Force:
    """Base class for single forces"""

    def __init__(self, *args, **kwargs) -> None:
        raise NotImplementedError

    def __call__(
        self, t: nt.Double, s: nt.CartesianState, fr: nt.Double = 0.0
    ) -> nt.CartesianStateDerivative:
        raise NotImplementedError


class ForceModel:
    """Complete force model to be passed to a the solver"""

    def __init__(self, *forces: Force) -> None:

        self.forces = forces
        self.n_forces = len(self.forces)
        return None

    def __call__(
        self, t: nt.Double, s: nt.CartesianState, fr: nt.Double = 0.0
    ) -> nt.CartesianStateDerivative:

        ds = self.forces[0](t, s, fr)
        if self.n_forces > 1:
            for i in range(1, self.n_forces):
                ds.add_acceleration(self.forces[i](t, s, fr))

        return ds



from typing import TypeVar, Generic
import numpy as np
from ..types.core import Vector, Double
from ..constants import miliarcsec

Frame = TypeVar("Frame", bound="ReferenceFrame")


class ReferenceFrame:
    """Base class for reference frames"""

    variables = {"r": "q1", "theta": "q2", "phi": "q3"}

    @classmethod
    def convert(
        cls, q1: Vector, q2: Vector, q3: Vector, frame: type[Frame]
    ) -> tuple[Vector, Vector, Vector]:
        raise NotImplementedError


class ICRF(ReferenceFrame):

    variables = {"r": "q1", "alpha": "q2", "delta": "q3"}

    dalpha0 = -14.6 * miliarcsec
    xi0 = -16.6170 * miliarcsec
    eta0 = -6.8192 * miliarcsec

    B = np.array(
        [
            [
                1.0 - 0.5 * (dalpha0 * dalpha0 + xi0 * xi0),
                dalpha0,
                -xi0,
            ],
            [
                -dalpha0 - eta0 * xi0,
                1.0 - 0.5 * (dalpha0 * dalpha0 + eta0 * eta0),
                -eta0,
            ],
            [
                xi0 - eta0 * dalpha0,
                eta0 + xi0 * dalpha0,
                1.0 - 0.5 * (xi0 * xi0 + eta0 * eta0),
            ],
        ]
    )

    @classmethod
    def convert(
        cls, q1: Vector, q2: Vector, q3: Vector, frame: type[Frame]
    ) -> tuple[Vector, Vector, Vector]:
        if frame == J2000:
            return np.dot(cls.B, np.array([q1, q2, q3]))
        else:
            raise NotImplementedError


class J2000(ReferenceFrame):

    variables = {"r": "q1", "theta": "q2", "phi": "q3"}

    @classmethod
    def convert(
        cls, q1: Vector, q2: Vector, q3: Vector, frame: type[Frame]
    ) -> tuple[Vector, Vector, Vector]:

        if frame == ICRF:
            return np.dot(ICRF.B.T, np.array([q1, q2, q3]))
        else:
            raise NotImplementedError



from ..types import KeplerianState, CartesianState, Vector
from ..constants import day
from . import core as ng

# TODO: ADD DOCSTRINGS


class StatePlot(ng.Mosaic):

    def __init__(self, setup: ng.PlotSetup = ng.PlotSetup()) -> None:

        super().__init__("ab;cd;ef", setup)

        base_setup = self.setup.copy()
        if self.setup.xlabel is None:
            base_setup.xlabel = "Days past initial epoch"

        subplot_setups = self.subplot_setup()

        self.q1_subplot = self.add_subplot(setup=subplot_setups[0])
        self.q2_subplot = self.add_subplot(setup=subplot_setups[1])
        self.q3_subplot = self.add_subplot(setup=subplot_setups[2])
        self.q4_subplot = self.add_subplot(setup=subplot_setups[3])
        self.q5_subplot = self.add_subplot(setup=subplot_setups[4])
        self.q6_subplot = self.add_subplot(setup=subplot_setups[5])

        return None

    def subplot_setup(self) -> list[ng.PlotSetup]:
        raise NotImplementedError

    def __exit__(self, exc_type, exc_value, traceback) -> None:

        self.q1_subplot.postprocess()
        self.q2_subplot.postprocess()
        self.q3_subplot.postprocess()
        self.q4_subplot.postprocess()
        self.q5_subplot.postprocess()
        self.q6_subplot.postprocess()

        super().__exit__(exc_type, exc_value, traceback)
        return None


class PlotState(StatePlot):

    def add_state(
        self,
        time: Vector,
        state: CartesianState | KeplerianState,
        fmt: str = "-",
        is_dt: bool = False,
    ) -> None:

        if not is_dt:
            time = (time - time[0]) / day

        self.q1_subplot.add_line(time, state.q1, fmt)
        self.q2_subplot.add_line(time, state.q2, fmt)

        if isinstance(state, KeplerianState):
            self.q3_subplot.add_line(time, state.i_deg, fmt)
            self.q4_subplot.add_line(time, state.raan_deg, fmt)
            self.q5_subplot.add_line(time, state.aop_deg, fmt)
            self.q6_subplot.add_line(time, state.ta_deg, fmt)
        else:
            self.q3_subplot.add_line(time, state.q3, fmt)
            self.q4_subplot.add_line(time, state.q4, fmt)
            self.q5_subplot.add_line(time, state.q5, fmt)
            self.q6_subplot.add_line(time, state.q6, fmt)

        return None


class CompareState(StatePlot):

    def compare_states[
        T: (CartesianState, KeplerianState)
    ](
        self,
        time: Vector,
        orbit: T,
        reference: T,
        is_dt: bool = False,
        fmt: str = "-",
    ) -> None:

        if not is_dt:
            time = (time - time[0]) / day

        ds = orbit - reference
        self.q1_subplot.add_line(time, ds.q1, fmt)
        self.q2_subplot.add_line(time, ds.q2, fmt)
        self.q3_subplot.add_line(time, ds.q3, fmt)
        self.q4_subplot.add_line(time, ds.q4, fmt)
        self.q5_subplot.add_line(time, ds.q5, fmt)
        self.q6_subplot.add_line(time, ds.q6, fmt)

        return None


class PlotKeplerianState(PlotState):
    """Plot components of a keplerian state vector"""

    def subplot_setup(self) -> list[ng.PlotSetup]:

        base_setup = self.setup.copy()

        a_setup = base_setup.copy()
        a_setup.ylabel = r"$a\ [m]$"

        e_setup = base_setup.copy()
        e_setup.ylabel = r"$e$"

        i_setup = base_setup.copy()
        i_setup.ylabel = r"$i\ [deg]$"

        aop_setup = base_setup.copy()
        aop_setup.ylabel = r"$\omega\ [deg]$"

        raan_setup = base_setup.copy()
        raan_setup.ylabel = r"$\Omega\ [deg]$"

        ta_setup = base_setup.copy()
        ta_setup.ylabel = r"$\theta\ [deg]$"

        return [a_setup, e_setup, i_setup, aop_setup, raan_setup, ta_setup]


class PlotCartesianState(PlotState):
    """Plot components of a cartesian state vector

    Parameters
    -----------
    setup : PlotSetup
        Plot setup
    """

    def subplot_setup(self) -> list[ng.PlotSetup]:

        base_setup = self.setup.copy()

        x_setup = base_setup.copy()
        x_setup.ylabel = r"$x\ [m]$"

        y_setup = base_setup.copy()
        y_setup.ylabel = r"$y\ [m]$"

        z_setup = base_setup.copy()
        z_setup.ylabel = r"$z\ [m]$"

        vx_setup = base_setup.copy()
        vx_setup.ylabel = r"$\dot x\ [m/s]$"

        vy_setup = base_setup.copy()
        vy_setup.ylabel = r"$\dot y\ [m/s]$"

        vz_setup = base_setup.copy()
        vz_setup.ylabel = r"$\dot z\ [m/s]$"

        return [x_setup, y_setup, z_setup, vx_setup, vy_setup, vz_setup]


class CompareCartesianStates(CompareState):
    """Plot difference between two sets of cartesian states"""

    def subplot_setup(self) -> list[ng.PlotSetup]:

        base_setup = self.setup.copy()

        dx_setup = base_setup.copy()
        dx_setup.ylabel = r"$\Delta x\ [m]$"

        dy_setup = base_setup.copy()
        dy_setup.ylabel = r"$\Delta y\ [m]$"

        dz_setup = base_setup.copy()
        dz_setup.ylabel = r"$\Delta z\ [m]$"

        dvx_setup = base_setup.copy()
        dvx_setup.ylabel = r"$\Delta \dot x\ [m/s]$"

        dvy_setup = base_setup.copy()
        dvy_setup.ylabel = r"$\Delta \dot y\ [m/s]$"

        dvz_setup = base_setup.copy()
        dvz_setup.ylabel = r"$\Delta \dot z\ [m/s]$"

        return [dx_setup, dy_setup, dz_setup, dvx_setup, dvy_setup, dvz_setup]


class CompareKeplerianStates(CompareState):
    """Plot difference between two sets of keplerian states"""

    def subplot_setup(self) -> list[ng.PlotSetup]:

        base_setup = self.setup.copy()

        da_setup = base_setup.copy()
        da_setup.ylabel = r"$\Delta a\ [m]$"

        de_setup = base_setup.copy()
        de_setup.ylabel = r"$\Delta e$"

        di_setup = base_setup.copy()
        di_setup.ylabel = r"$\Delta i\ [rad]$"

        daop_setup = base_setup.copy()
        daop_setup.ylabel = r"$\Delta \omega\ [rad]$"

        draan_setup = base_setup.copy()
        draan_setup.ylabel = r"$\Delta \Omega\ [rad]$"

        dta_setup = base_setup.copy()
        dta_setup.ylabel = r"$\Delta \theta\ [rad]$"

        return [da_setup, de_setup, di_setup, daop_setup, draan_setup, dta_setup]


class PlotOrbit(ng.Base3D):

    def add_orbit(
        self,
        state: CartesianState,
        fmt: str = "-",
        width: float | None = None,
        markersize: float | None = None,
        color: str | None = None,
        alpha: float = 1.0,
        label: str | None = None,
        axis: str = "left",
    ) -> None:

        self.add_line(
            state.x, state.y, state.z, fmt, width, markersize, color, alpha, label, axis
        )

        return None



from .base import CartesianPropagator


class EulerPropagator(CartesianPropagator):

    def __init__(self, fun, t0, s0, tend, dt) -> None:

        super().__init__(fun, t0, s0, tend, dt)

        return None

    def _step_impl(self):

        new_time_int, new_time_fr = self._update_epoch()
        h_int, h_fr = self._get_steps()

        f = self.model(self.now_int, self.s, fr=self.now_fr)
        new_s = self.s + f * h_int + f * h_fr

        # new_s = self.s + self.fun(self.now_int, self.s, fr=self.now_fr) * h

        self.now_int = new_time_int
        self.now_fr = new_time_fr
        self.s = new_s

        return True



from .base import CartesianPropagator
from .euler import EulerPropagator
from .rk import RungeKuttaPropagator, SimpleRK4

__all__ = [
    "CartesianPropagator",
    "EulerPropagator",
    "RungeKuttaPropagator",
    "SimpleRK4",
]



"""
Catalog
==============

.. module:: nastro.catalog

Base class
----------------
.. autosummary::
    :toctree: generated/
    :nosignatures:

    CelestialBody

Celestial bodies
------------------

=============================== ================================================
Planet                          Moons
=============================== ================================================
Sun                             None
Mercury                         None
Venus                           None
Earth                           Moon
Mars                            Phobos, Deimos
Jupiter                         Io, Europa, Ganymede, Callisto
Saturn                          Mimas, Enceladus, Tethys, Dione, Rhea, Titan
Uranus                          Ariel, Umbriel, Titania, Oberon, Miranda
Neptune                         Triton
Pluto                           Charon
=============================== ================================================
"""

from .types.core import Double
from . import constants as nc


class CelestialBody:
    """Properties of celestial bodies

    Sources:
    - GM data for planets: https://doi.org/10.3847/1538-3881/abd414
    - Radii: https://doi.org/10.1007/s10569-017-9805-5
    - Earth data: Horizons geophysical data [March 2024]
    - Physical properties of the Moon: https://doi.org/10.1007/s11214-019-0613-y

    Properties
    ----------
    name : str
        Name of the celestial body
    R : Double
        Mean radius [m]
    Rp : Double
        Polar radius [m]
    Re : Double
        Equatorial radius [m]
    rho : Double
        Mean density [kg/m^3]
    mu : Double
        Gravitational parameter [m^3/s^2]
    j2 : Double
        Second zonal harmonic
    j3 : Double
        Third zonal harmonic
    T : Double
        Orbital period [s]
    a : Double
        Semi-major axis [m]
    mass : Double
        Mass [kg]
    volume : Double
        Volume [m^3]
    inertia_factor : Double
        Inertia factor (I/MR^2)
    """

    _attributes = [
        "R",
        "Rp",
        "Re",
        "mu",
        "j2",
        "j3",
        "T",
        "a",
        "mass",
        "volume",
        "rho",
        "inertia_factor",
    ]

    def __init__(
        self,
        name: str,
        R: Double | None = None,
        Rp: Double | None = None,
        Re: Double | None = None,
        mu: Double | None = None,
        j2: Double | None = None,
        j3: Double | None = None,
        T: Double | None = None,
        a: Double | None = None,
        inertia_factor: Double | None = None,
    ) -> None:
        self.name = name
        self.__R = R
        self.__Rp = Rp
        self.__Re = Re
        self.__mu = mu
        self.__j2 = j2
        self.__j3 = j3
        self.__T = T
        self.__a = a
        self.__inertia_factor = inertia_factor

        self.__mass = None if mu is None else mu / nc.G
        self.__volume = None if R is None else (4.0 / 3.0) * nc.pi * (R**3)
        self.__rho = (
            None
            if self.__mass is None or self.__volume is None
            else self.__mass / self.__volume
        )

    def __getattr__(self, name: str) -> Double:

        if name in self._attributes:

            val = self.__getattribute__(f"_CelestialBody__{name}")
            if val is None:
                raise AttributeError(f"{name} is not specified for '{self.name}'")
            return val

        raise AttributeError(f"'{self.name}' has no attribute '{name}'")

    def __repr__(self) -> str:

        out = f"Physical properties: {self.name}\n"
        out += "-" * len(out) + "\n"
        for attr in self._attributes:
            try:
                val = getattr(self, attr)
                out += f"{attr}:\t{val:e}\n"
            except AttributeError:
                continue
        return out


Sun = CelestialBody(
    "Sun",
    mu=1.3271244004127942e20,
    Re=695700e3,
)

Mercury = CelestialBody(
    "Mercury",
    R=2.4394e6,
    Re=2440.53e3,
    Rp=2438.26e3,
    mu=2.2031868551400003e13,
)

Venus = CelestialBody(
    "Venus",
    R=6051.8e3,
    Re=6051.8e3,
    Rp=6051.8e3,
    mu=3.2485859200000000e14,
    a=108.210e9,
)

Earth = CelestialBody(
    "Earth",
    R=6371.0084e3,
    Re=6378.1366e3,
    Rp=6356.7519e3,
    mu=3.9860043550702266e14,
    j2=1.08262545e-3,
    T=3.155814950400000e07,
    a=149.6e9,
    inertia_factor=0.3307,
)

Mars = CelestialBody(
    "Mars",
    R=3389.50e3,
    Re=3396.19e3,
    mu=4.282837362069909e13,
)

Jupiter = CelestialBody(
    "Jupiter",
    R=69911e3,
    Re=71492e3,
    Rp=66854e3,
    mu=1.266865341960128e17,
    j2=14.696572e-3,
    j3=-0.042e-6,
)

Saturn = CelestialBody(
    "Saturn",
    R=58232e3,
    Re=60268e3,
    Rp=54364e3,
    mu=3.793120623436167e16,
)

Uranus = CelestialBody(
    "Uranus",
    R=25362e3,
    Re=25559e3,
    Rp=24973e3,
    mu=5.793951256527211e15,
)

Neptune = CelestialBody(
    "Neptune",
    R=24622e3,
    Re=24764e3,
    Rp=24341e3,
    mu=6.835103145462294e15,
)


# Moons
Moon = CelestialBody(
    "Moon",
    R=1.7374e6,
    mu=4.9028001184575496e12,
    inertia_factor=0.3931,
)

Phobos = CelestialBody(
    "Phobos",
    R=11.08e3,
    mu=7.087546066894452e5,
)

Deimos = CelestialBody(
    "Deimos",
    R=6.2e3,
    mu=9.615569648120313e4,
)

Io = CelestialBody(
    "Io",
    R=1821.49e3,
    mu=5.959924010272514e12,
    T=1.5292800000000e05,
)

Europa = CelestialBody(
    "Europa",
    R=1560.8e3,
    mu=3.202739815114734e12,
    T=3.0672000000000e05,
)

Ganymede = CelestialBody(
    "Ganymede",
    R=2631.2e3,
    mu=9.887819980080976e12,
    T=6.1819200000000e05,
)

Callisto = CelestialBody(
    "Callisto",
    R=2410.3e3,
    mu=7.179304867611079e12,
    T=1.4421024000000e06,
)

Mimas = CelestialBody(
    "Mimas",
    R=198.2e3,
    mu=2.503488768152587e9,
)

Enceladus = CelestialBody(
    "Enceladus",
    R=252.1e3,
    mu=7.210366688598896e9,
)

Tethys = CelestialBody(
    "Tethys",
    R=531.0e3,
    mu=4.121352885489587e10,
)

Dione = CelestialBody(
    "Dione",
    R=561.4e3,
    mu=7.311607172482067e10,
)

Rhea = CelestialBody(
    "Rhea",
    R=763.5e3,
    mu=1.539417519146563e11,
)

Titan = CelestialBody(
    "Titan",
    R=2575.0e3,
    mu=8.978137095521046e12,
)

Ariel = CelestialBody(
    "Ariel",
    R=578.9e3,
    mu=8.346344431770477e10,
)

Umbriel = CelestialBody(
    "Umbriel",
    R=584.7e3,
    mu=8.509338094489388e10,
)

Titania = CelestialBody(
    "Titania",
    R=788.9e3,
    mu=2.269437003741248e11,
)

Oberon = CelestialBody(
    "Oberon",
    R=761.4e3,
    mu=2.053234302535623e11,
)

Miranda = CelestialBody(
    "Miranda",
    R=235.8e3,
    mu=4.319516899232100e9,
)

Triton = CelestialBody(
    "Triton",
    R=1352.6e3,
    mu=1.428495462910464e12,
)

# Dwarf planets
Pluto = CelestialBody(
    "Pluto",
    R=1188.3e3,
    mu=8.696138177608748e11,
)

Charon = CelestialBody(
    "Charon",
    R=606.0e3,
    mu=1.058799888601881e11,
)

Ceres = CelestialBody(
    "Ceres",
    R=470e3,
)



import numpy as np
from typing import Self, Any, Iterator
from pathlib import Path
from .core import Double, Vector, Scalar, is_double, is_vector, is_scalar
import traceback
from ..constants import halfpi, pi, twopi


class GenericState[U: (Double, Vector)]:
    """Base class for state vectors

    Implements operations and basic functionality that is common to all types
    of state vectors and implements the containers that are used internally
    to manipulate their components. The elements q1 to q6 are representations
    of the six components of a state vector as numpy arrays.

    This class is not meant to be instantiated directly, but to be subclassed
    to create specific types of state vectors: cartesian, keplerian, etc.
    """

    __slots__ = ("q1", "q2", "q3", "q4", "q5", "q6")
    properties = {
        "q1": "q1",
        "q2": "q2",
        "q3": "q3",
        "q4": "q4",
        "q5": "q5",
        "q6": "q6",
    }
    angles: dict[str, tuple[Double, Double]] = {}

    def __init__(
        self,
        q1: U,
        q2: U,
        q3: U,
        q4: U,
        q5: U,
        q6: U,
        deg: bool = False,
        wrap: bool = False,
    ) -> None:

        if is_double(q1):
            _input = np.array([q1, q2, q3, q4, q5, q6], dtype=np.float64)[:, None]
            assert len(_input.shape) == 2
            assert _input.shape == (6, 1)
        elif is_vector(q1):
            try:
                _input = np.array([q1, q2, q3, q4, q5, q6], dtype=np.float64)
                assert len(_input.shape) == 2
                assert _input.shape[0] == 6
            except ValueError or AssertionError:
                raise ValueError(
                    f"Failed to create {self.__class__.__name__}. "
                    "Components have different sizes."
                )
        else:
            raise TypeError(
                f"Failed to create {self.__class__.__name__}. "
                "Components must be floating point numbers or vectors."
            )

        # Initialize generic state components
        self.q1 = np.array(q1, dtype=np.float64, ndmin=1)
        self.q2 = np.array(q2, dtype=np.float64, ndmin=1)
        self.q3 = np.array(q3, dtype=np.float64, ndmin=1)
        self.q4 = np.array(q4, dtype=np.float64, ndmin=1)
        self.q5 = np.array(q5, dtype=np.float64, ndmin=1)
        self.q6 = np.array(q6, dtype=np.float64, ndmin=1)

        if len(self.angles.keys()) == 0:
            return None

        if deg:
            for key in self.angles.keys():
                setattr(
                    self,
                    self.properties[key],
                    np.deg2rad(getattr(self, self.properties[key])),
                )
        if wrap:
            for key, limits in self.angles.items():
                setattr(
                    self,
                    self.properties[key],
                    self.__wrap_angle(getattr(self, self.properties[key]), limits),
                )

        return None

    @property
    def size(self) -> int:
        """Number of state vectors in time series"""
        return self.q1.size

    @property
    def scalar(self) -> bool:
        """True for a single state vector"""
        return self.q1.size == 1

    @property
    def asarray(self) -> Vector:
        """Time series of state vectors as a (6, N) numpy array"""
        out = np.array(
            [self.q1, self.q2, self.q3, self.q4, self.q5, self.q6],
            dtype=np.float64,
        )
        return out

    @staticmethod
    def __wrap_angle(angle: U, limits: tuple[Double, Double] = (0.0, twopi)) -> U:
        """Wrap angle to limits"""

        low, high = limits
        angle %= twopi
        if low == 0 and high == twopi:
            return angle

        if low == 0 and high == pi:
            return angle % pi

        out = angle - ((angle + pi) // twopi) * twopi
        if low == -pi and high == pi:
            return out

        positive = (out * (out < halfpi) + (pi - out) * (out >= halfpi)) * (out >= 0)
        negative = (out * (out > -halfpi) - (pi + out) * (out <= -halfpi)) * (out < 0)

        if low == -halfpi and high == halfpi:
            return positive + negative

        raise ValueError("Failed to wrap angle: Invalid limits.")

    def __getattr__(self, name: str) -> U:

        if name in self.properties:
            qi = getattr(self, self.properties[name])
            return qi[0] if self.scalar else qi

        if name[-4:] == "_deg" and name[:-4] in self.angles:
            key = name[:-4]
            return np.rad2deg(getattr(self, key))

        raise AttributeError(f"{name} is not a property of {self.__class__.__name__}! ")

    def __getitem__(self, index: int | slice) -> Self:

        return type(self)(*self.asarray[:, index])

    def __iter__(self) -> Iterator[Self]:

        if self.scalar:
            return iter(
                [self.q1[0], self.q2[0], self.q3[0], self.q4[0], self.q5[0], self.q6[0]]
            )
        return iter([self.__getitem__(idx) for idx in range(self.size)])

    def __repr__(self) -> str:

        out = f"{self.__class__.__name__}\n"
        out += "-" * len(out) + "\n"
        for key in self.properties:
            out += f"{key}: {getattr(self, key)}\n"
        return out

    def __add__(self, other: Self | Scalar) -> Self:

        if is_scalar(other):
            return type(self)(*(self.asarray + other))
        elif isinstance(other, self.__class__):
            return type(self)(*(self.asarray + other.asarray))
        else:
            raise TypeError(
                "Failed to perform addition! Check documentation for supported types."
            )

    def __sub__(self, other: Self | Scalar) -> Self:

        if is_scalar(other):
            return type(self)(*(self.asarray - other))
        elif isinstance(other, self.__class__):
            return type(self)(*(self.asarray - other.asarray))
        else:
            raise TypeError(
                "Failed to perform subtraction! "
                "Check documentation for supported types."
            )

    def __mul__(self, other: Self | Scalar) -> Self:

        if is_scalar(other):
            return type(self)(*(self.asarray * other))
        elif isinstance(other, self.__class__):
            return type(self)(*(self.asarray * other.asarray))
        else:
            raise TypeError(
                "Failed to perform multiplication! "
                "Check documentation for supported types."
            )

    def __truediv__(self, other: Self | Scalar) -> Self:

        if is_scalar(other):
            return type(self)(*(self.asarray / other))
        elif isinstance(other, self.__class__):
            return type(self)(*(self.asarray / other.asarray))
        else:
            raise TypeError(
                "Failed to perform division! "
                "Check documentation for supported types."
            )

    def __eq__(self, other: object) -> bool:

        if not isinstance(other, self.__class__):
            raise TypeError("Comparison is only supported for state vectors")

        return np.allclose(self.asarray, other.asarray, atol=1e-15, rtol=0.0)

    def append(self, other: Self | Scalar | Vector) -> None:
        """Add state vector to time series"""

        if is_scalar(other):
            values = np.ones((6, 1)) * other
        elif is_vector(other):
            if other.shape[0] != 6:
                raise ValueError(
                    "Failed to append to state vector. Input must be a (6, N) array."
                )
            if len(other.shape) == 1:
                other = other[:, None]
            values = other
        elif isinstance(other, self.__class__):
            values = other.asarray
        else:
            raise TypeError(
                "Failed to append to state vector. "
                "Check documentation for supported types."
            )

        self.q1 = np.append(self.q1, values[0], axis=0)
        self.q2 = np.append(self.q2, values[1], axis=0)
        self.q3 = np.append(self.q3, values[2], axis=0)
        self.q4 = np.append(self.q4, values[3], axis=0)
        self.q5 = np.append(self.q5, values[4], axis=0)
        self.q6 = np.append(self.q6, values[5], axis=0)
        return None

    def copy(self) -> Self:
        """Return copy of state vector"""
        return type(self)(*self.asarray)

    # IO
    def save(self, path: str | Path, relative: bool = True) -> Path:

        # Add extension if missing
        path = Path(path)
        if not path.suffix:
            path = path.with_suffix(".npy")

        # Generate absolute path
        cwd = Path("/")
        if relative:
            cwd = Path(traceback.extract_stack()[-2].filename).parent
        path = cwd / path

        # Save output
        path.parent.mkdir(parents=True, exist_ok=True)
        np.save(path, self.asarray)

        return path

    @classmethod
    def load(cls, path: str | Path, relative: bool = True) -> Self:

        # Add extension if missing
        path = Path(path)
        if not path.suffix:
            path = path.with_suffix(".npy")

        # Generate absolute path
        cwd = Path("/")
        if relative:
            cwd = Path(traceback.extract_stack()[-2].filename).parent
        path = cwd / path

        # Load output
        if path.suffix == ".npy":
            return cls(*np.load(path))
        elif path.suffix == ".dat":
            data = np.loadtxt(path).T
            if data.shape[0] == 6:
                return cls(*data)
            elif data.shape[0] >= 7:
                return cls(*data[1:7])
            else:
                raise ValueError("Failed to load state vector. Invalid data shape.")
        else:
            raise ValueError("Failed to load state vector. Invalid file extension.")

    @classmethod
    def from_tudat(cls, state_history: dict[Double, list[Double]]) -> Self:
        """Generate state vector from Tudat state history

        :param state_history: Dictionary with epochs and state components
        """
        return cls(*np.array(list(state_history.values())).T[:6])

    # Frame conversions
    def transform(self) -> Self:
        """Transform state to a different reference frame"""
        raise NotImplementedError("Frame transformations are not implemented yet")


class CartesianState[U: (Double, Vector)](GenericState[U]):
    """Cartesian state vector

    Single state or a time series of states of a body in a three-dimensional
    space in terms of cartesian coordinates. Each component of the state vector
    can be a floating point number (single state) or a numpy array of them (time
    series of states).

    :param x: X component of the position vector [m]
    :param y: Y component of the position vector [m]
    :param z: Z component of the position vector [m]
    :param dx: X component of the velocity vector [m/s]
    :param dy: Y component of the velocity vector [m/s]
    :param dz: Z component of the velocity vector [m/s]
    """

    properties = {
        "x": "q1",
        "y": "q2",
        "z": "q3",
        "dx": "q4",
        "dy": "q5",
        "dz": "q6",
    }

    def __init__(self, x: U, y: U, z: U, dx: U, dy: U, dz: U) -> None:
        super().__init__(x, y, z, dx, dy, dz)

    @property
    def r_vec(self) -> Vector:
        """Cartesian position vector as numpy array"""
        return np.array([self.x, self.y, self.z])

    @property
    def r_mag(self) -> U:
        """Magnitude of the position vector"""
        return np.linalg.norm(self.r_vec, axis=0)

    @property
    def v_vec(self) -> Vector:
        """Cartesian velocity vector as numpy array"""
        return np.array([self.dx, self.dy, self.dz])

    @property
    def v_mag(self) -> U:
        """Magnitude of the velocity vector"""
        return np.linalg.norm(self.v_vec, axis=0)

    def to_keplerian(self, mu: Double) -> "KeplerianState":
        """Conversion to keplerian state vector

        Algorithm: Practical astrodynamics slides

        :param mu: Gravitational parameter of central body [m^3/s^2]
        """
        # TODO: Ensure that loss of accuracy in np.sum is not relevant
        # TODO: Only allow conversion if frame is valid

        # Semi-major axis
        a = 1.0 / ((2.0 / self.r_mag) - (self.v_mag * self.v_mag / mu))

        # Angular momentum
        h_vec = np.cross(self.r_vec, self.v_vec, axis=0)
        h = np.linalg.norm(h_vec, axis=0)

        # Eccentricity
        e_vec = (np.cross(self.v_vec, h_vec, axis=0) / mu) - (self.r_vec / self.r_mag)
        e = np.linalg.norm(e_vec, axis=0)
        e_uvec = e_vec / e

        # Inclination
        inc = np.arccos(h_vec[2] / h)

        # N vector
        _z_vec = np.array(
            [np.zeros_like(self.x), np.zeros_like(self.y), np.ones_like(self.z)]
        )
        N_vec = np.cross(_z_vec, h_vec, axis=0)
        Nxy = np.sqrt(N_vec[0] * N_vec[0] + N_vec[1] * N_vec[1])
        N_uvec = N_vec / Nxy

        # RAAN
        raan = np.arctan2(N_vec[1] / Nxy, N_vec[0] / Nxy)

        # Argument of periapsis
        sign_aop_condition = np.sum(np.cross(N_uvec, e_vec, axis=0) * h_vec, axis=0) > 0
        sign_aop = 2 * sign_aop_condition - 1
        aop = sign_aop * np.arccos(np.sum(e_uvec * N_uvec, axis=0))

        # True anomaly
        sign_ta_condition = (
            np.sum(np.cross(e_vec, self.r_vec, axis=0) * h_vec, axis=0) > 0
        )
        sign_ta = 2 * sign_ta_condition - 1
        ta = sign_ta * np.arccos(np.sum(self.r_vec * e_uvec / self.r_mag, axis=0))

        return KeplerianState(a, e, inc, raan, aop, ta, deg=False)

    def to_spherical(self):
        raise NotImplementedError(
            "Conversion to spherical state is not implemented yet."
        )


class CartesianStateDerivative[U: (Double, Vector)](GenericState[U]):
    """Derivative of cartesian state

    Derivative of a single state of a time series of states of a body in a
    three-dimensional space in terms of cartesian coordinates. Each component
    can be a floating point number (single state) or a numpy array of them
    (time series of states).

    :param dx: X component of the velocity vector [m/s]
    :param dy: Y component of the velocity vector [m/s]
    :param dz: Z component of the velocity vector [m/s]
    :param ddx: X component of the acceleration vector [m/s^2]
    :param ddy: Y component of the acceleration vector [m/s^2]
    :param ddz: Z component of the acceleration vector [m/s^2]
    """

    properties = {
        "dx": "q1",
        "dy": "q2",
        "dz": "q3",
        "ddx": "q4",
        "ddy": "q5",
        "ddz": "q6",
    }

    def __init__(self, dx: U, dy: U, dz: U, ddx: U, ddy: U, ddz: U) -> None:
        super().__init__(dx, dy, dz, ddx, ddy, ddz)
        return None

    @property
    def v_vec(self) -> Vector:
        """Cartesian velocity vector as numpy array"""
        return np.array([self.dx, self.dy, self.dz])

    @property
    def a_vec(self) -> Vector:
        """Cartesian acceleration vector as numpy array"""
        return np.array([self.ddx, self.ddy, self.ddz])

    @property
    def v_mag(self) -> U:
        """Magnitude of the velocity vector"""
        return np.linalg.norm(self.v_vec, axis=0)

    @property
    def a_mag(self) -> U:
        """Magnitude of the acceleration vector"""
        return np.linalg.norm(self.a_vec, axis=0)

    def times_dt(self, dt: Double) -> CartesianState:
        """Change in cartesian state over interval dt

        Returns the change of a cartesian state with this object as derivative
        over a time interval dt. The main use case are numerical integrators,
        in which the state is updated by multiplying its derivative by the
        time step.

        :param dt: Time step
        :return: Change in cartesian elements over interval dt
        """

        return CartesianState(*(self.asarray * dt))

    def add_acceleration(self, other: Self) -> None:
        """Update the acceleration of the cartesian state derivative

        Update the acceleration of this object by adding the components of the
        acceleration vector of other cartesian state derivative. The use case is
        to add up all the accelerations acting on a body during orbit propagation.
        Since each acceleration is defined by a cartesian state derivative,
        adding them up directly would also update the velocity, which is incorrect.

        :param other: Cartesian state derivative with acceleration to add
        """

        if not other.scalar:
            raise NotImplementedError(
                "Failed to add acceleration. Vector input not supported."
            )

        # q4, q5 and q6 are the components of a general state vector that
        # represent the acceleration in CartesianStateDerivative
        self.q4 += other.q4
        self.q5 += other.q5
        self.q6 += other.q6

        return None


class KeplerianState[U: (Double, Vector)](GenericState[U]):
    """Keplerian state vector

    Single state of a time series of states of a body in a three-dimensional
    space in terms of classical keplerian elements. Each component of the state
    vector can be a floating point number (single state) or a numpy array of them
    (time series of states).

    Angles might be given in degrees or radians, but are always converted to
    radians during initialization. For each angle, the method KeplerianState.angle
    will return the value in radians, while KeplerianState.angle_deg will return
    it in degrees.

    Angles can be wrapped if specified by the wrap parameter, which is activated
    by default. Wrapping will limit inclinations to the range [0, pi) and the
    rest of the angles to [0, 2pi).

    :param a: Semi-major axis [m]
    :param e: Eccentricity
    :param i: Inclination [rad]
    :param raan: Right ascension of the ascending node [rad]
    :param aop: Argument of periapsis [rad]
    :param ta: True anomaly [rad]
    :param deg: Angles are given in degrees (True) or radians (False)
    :param wrap: Wrap angles (True) or not (False)
    """

    properties = {
        "a": "q1",
        "e": "q2",
        "i": "q3",
        "raan": "q4",
        "aop": "q5",
        "ta": "q6",
    }

    angles = {
        "i": (0.0, pi),
        "raan": (0.0, twopi),
        "aop": (0.0, twopi),
        "ta": (0.0, twopi),
    }

    def __init__(
        self,
        a: U,
        e: U,
        i: U,
        raan: U,
        aop: U,
        ta: U,
        deg: bool = False,
        wrap: bool = True,
    ) -> None:
        super().__init__(a, e, i, raan, aop, ta, deg=deg, wrap=wrap)
        return None

    @property
    def E(self) -> U:
        """Eccentric anomaly"""
        if np.any(self.e >= 1.0) or np.any(self.e < 0.0):
            raise NotImplementedError(
                "Failed to compute eccentric anomaly. "
                "Only circular and elliptical orbits are supported"
            )
        out: Any = 2.0 * np.arctan2(
            np.sqrt(1.0 - self.e) * np.sin(0.5 * self.ta),
            np.sqrt(1.0 + self.e) * np.cos(0.5 * self.ta),
        )
        return out

    @property
    def M(self) -> U:
        """Mean anomaly"""
        out: Any = self.E - self.e * np.sin(self.E)
        return out

    @property
    def T(self) -> U:
        """Orbital period"""
        out: Any = 2.0 * np.pi * np.sqrt(self.a**3)
        return out

    def __sub__(self, other: Self | Scalar) -> Self:
        print(
            "WARNING: Keplerian state subtraction is not properly tested\n"
            "It might produce unexpected results"
        )

        if not isinstance(other, self.__class__):
            raise NotImplementedError(
                "Subtraction from KeplerianState is only supported "
                "for KeplerianState objects."
            )

        da: Any = self.q1 - other.q1
        de: Any = self.q2 - other.q2
        di: Any = np.unwrap(self.q3, period=pi) - np.unwrap(other.q3, period=pi)
        draan: Any = np.unwrap(self.q4, period=twopi) - np.unwrap(
            other.q4, period=twopi
        )
        daop: Any = np.unwrap(self.q5, period=twopi) - np.unwrap(other.q5, period=twopi)
        dta: Any = np.unwrap(self.q6, period=twopi) - np.unwrap(other.q6, period=twopi)

        return type(self)(da, de, di, draan, daop, dta, wrap=False)

    def to_cartesian(self, mu: Double) -> "CartesianState":
        """Conversion to cartesian state vector

        Algorithm: Practical astrodynamics slides

        :param mu: Gravitational parameter of central body [m^3/s^2]
        """

        # Auxiliary trigonometric relations
        cos_Omega = np.cos(self.raan)
        sin_Omega = np.sin(self.raan)
        cos_omega = np.cos(self.aop)
        sin_omega = np.sin(self.aop)
        cos_i = np.cos(self.i)
        sin_i = np.sin(self.i)
        cos_theta = np.cos(self.ta)
        sin_theta = np.sin(self.ta)

        l1 = cos_Omega * cos_omega - sin_Omega * sin_omega * cos_i
        l2 = -cos_Omega * sin_omega - sin_Omega * cos_omega * cos_i
        m1 = sin_Omega * cos_omega + cos_Omega * sin_omega * cos_i
        m2 = -sin_Omega * sin_omega + cos_Omega * cos_omega * cos_i
        n1 = sin_omega * sin_i
        n2 = cos_omega * sin_i

        # Orbital radius
        r = self.a * (1.0 - self.e * self.e) / (1.0 + self.e * cos_theta)

        # Position in orbital plane
        xi = r * cos_theta
        eta = r * sin_theta

        # Position in 3D space
        x = l1 * xi + l2 * eta
        y = m1 * xi + m2 * eta
        z = n1 * xi + n2 * eta

        # Angular momentum
        H = np.sqrt(mu * self.a * (1.0 - self.e * self.e))

        # Velocity
        common = mu / H
        e_cos_theta = self.e + cos_theta

        dx = common * (l2 * e_cos_theta - l1 * sin_theta)
        dy = common * (m2 * e_cos_theta - m1 * sin_theta)
        dz = common * (n2 * e_cos_theta - n1 * sin_theta)

        return CartesianState(x, y, z, dx, dy, dz)


class KeplerianStateDerivative[U: (Double, Vector)](GenericState[U]):
    """Derivative of keplerian state

    Derivative of a single state of a time series of states of a body in a
    three-dimensional space in terms of classical keplerian elements. Each
    component can be a floating point number (single state) or a numpy array
    of them (time series of states).

    :param da: Rate of change of semi-major axis [m/s]
    :param de: Rate of change of eccentricity [1/s]
    :param di: Rate of change of inclination [rad/s]
    :param draan: Rate of change of right ascension of the ascending node [rad/s]
    :param daop: Rate of change of argument of periapsis [rad/s]
    :param dta: Rate of change of true anomaly [rad/s]
    """

    properties = {
        "da": "q1",
        "de": "q2",
        "di": "q3",
        "draan": "q4",
        "daop": "q5",
        "dta": "q6",
    }

    def __init__(self, da: U, de: U, di: U, draan: U, daop: U, dta: U) -> None:
        super().__init__(da, de, di, draan, daop, dta)
        return None

    def times_dt(self, dt: Double) -> KeplerianState:
        """Change in keplerian state over interval dt

        Returns the change of a keplerian state with this object as derivative
        over a time interval dt. The main use case are numerical integrators,
        in which the state is updated by multiplying its derivative by the
        time step.

        :param dt: Time step
        :return: Change in keplerian elements over interval dt
        """

        return KeplerianState(*(self.asarray * dt))


class SphericalState[U: (Double, Vector)](GenericState):

    def __init__(
        self,
        rho: U,
        theta: U,
        phi: U,
        drho: U,
        dtheta: U,
        dphi: U,
        deg: bool = False,
        wrap: bool = True,
    ) -> None:
        raise NotImplementedError("SphericalState is not implemented yet.")

    def __add__(self, other: object) -> Self:
        raise NotImplementedError(
            "Addition is still not supported for spherical state vectors"
        )

    def __sub__(self, other: object) -> Self:
        raise NotImplementedError(
            "Subtraction is still not supported for spherical state vectors"
        )

    def transform(self) -> Self:
        raise NotImplementedError(
            "Frame transformations are not implemented yet for spherical state vectors"
        )

    def to_cartesian(self) -> "CartesianState":
        """Conversion from spherical to cartesian state"""
        raise NotImplementedError(
            "Conversion from spherical to cartesian state is not implemented yet."
        )

    def to_keplerian(self, mu: Double) -> "KeplerianState":
        """Conversion from spherical to keplerian state"""
        raise NotImplementedError(
            "Conversion from spherical to keplerian state is not implemented yet."
        )


class CartesianPosition[U: (Double, Vector)](GenericState[U]):
    """Three dimensional position vector in cartesian coordinates

    Components might be floating point numbers (single vector) or numpy arrays
    of them (time series of vectors).

    :param x: X coordinate [m]
    :param y: Y coordinate [m]
    :param z: Z coordinate [m]
    """

    properties = {"x": "q1", "y": "q2", "z": "q3"}

    def __init__(self, x: U, y: U, z: U, *_) -> None:
        super().__init__(x, y, z, 0.0 * x, 0.0 * y, 0.0 * z)
        return None

    @property
    def asarray(self) -> Vector:
        """Time series of position vectors as a (3, N) numpy array"""
        return np.array([self.q1, self.q2, self.q3], dtype=np.float64)


class CartesianVelocity[U: (Double, Vector)](GenericState[U]):
    """Three dimensional velocity vector in cartesian coordinates

    Components might be floating point numbers (single vector) or numpy arrays
    of them (time series of vectors).

    :param dx: X component of the velocity vector [m/s]
    :param dy: Y component of the velocity vector [m/s]
    :param dz: Z component of the velocity vector [m/s]
    """

    properties = {"dx": "q4", "dy": "q5", "dz": "q6"}

    def __init__(self, dx: U, dy: U, dz: U, *_) -> None:
        super().__init__(0.0 * dx, 0.0 * dy, 0.0 * dz, dx, dy, dz)
        return None

    @property
    def asarray(self) -> Vector:
        """Time series of velocity vectors as a (3, N) numpy array"""
        return np.array([self.q4, self.q5, self.q6], dtype=np.float64)


class SphericalPosition[U: (Double, Vector)](GenericState[U]):

    def __init__(self) -> None:
        raise NotImplementedError("SphericalPosition is not implemented yet.")


# class _CartesianState[U: (Double, Vector)](GenericState):
#     """Cartesian state vector

#     Single state or a time series of states of a body in a three-dimensional
#     space in terms of cartesian coordinates. Each component of the state vector
#     can be a floating point number (single state) or a numpy array of them (time
#     series of states).

#     :param x: X component of the position vector [m]
#     :param y: Y component of the position vector [m]
#     :param z: Z component of the position vector [m]
#     :param dx: X component of the velocity vector [m/s]
#     :param dy: Y component of the velocity vector [m/s]
#     :param dz: Z component of the velocity vector [m/s]
#     """

#     def __init__(self, x: U, y: U, z: U, dx: U, dy: U, dz: U) -> None:
#         super().__init__(x, y, z, dx, dy, dz)
#         return None

#     @property
#     def x(self) -> U:
#         return self.q1[0] if self.scalar else self.q1

#     @property
#     def y(self) -> U:
#         return self.q2[0] if self.scalar else self.q2

#     @property
#     def z(self) -> U:
#         return self.q3[0] if self.scalar else self.q3

#     @property
#     def dx(self) -> U:
#         return self.q4[0] if self.scalar else self.q4

#     @property
#     def dy(self) -> U:
#         return self.q5[0] if self.scalar else self.q5

#     @property
#     def dz(self) -> U:
#         return self.q6[0] if self.scalar else self.q6

#     @property
#     def r_vec(self) -> Vector:
#         return np.array([self.x, self.y, self.z])

#     @property
#     def v_vec(self) -> Vector:
#         return np.array([self.dx, self.dy, self.dz])

#     @property
#     def r_mag(self) -> U:
#         return np.linalg.norm(self.r_vec, axis=0)

#     @property
#     def v_mag(self) -> U:
#         return np.linalg.norm(self.v_vec, axis=0)

#     def __repr__(self) -> str:
#         return (
#             "Cartesian state vector\n"
#             "----------------------\n"
#             f"x: {self.x}\n"
#             f"y: {self.y}\n"
#             f"z: {self.z}\n"
#             f"dx: {self.dx}\n"
#             f"dy: {self.dy}\n"
#             f"dz: {self.dz}"
#         )

#     def as_keplerian(self, mu: Double) -> "KeplerianState":
#         """Conversion to keplerian state vector

#         Algorithm: Practical astrodynamics slides

#         :param mu: Gravitational parameter of central body [m^3/s^2]
#         """
#         # TODO: Ensure that loss of accuracy in np.sum is not relevant

#         # Semi-major axis
#         a = 1.0 / ((2.0 / self.r_mag) - (self.v_mag * self.v_mag / mu))

#         # Angular momentum
#         h_vec = np.cross(self.r_vec, self.v_vec, axis=0)
#         h = np.linalg.norm(h_vec, axis=0)

#         # Eccentricity
#         e_vec = (np.cross(self.v_vec, h_vec, axis=0) / mu) - (self.r_vec / self.r_mag)
#         e = np.linalg.norm(e_vec, axis=0)
#         e_uvec = e_vec / e

#         # Inclination
#         inc = np.arccos(h_vec[2] / h)

#         # N vector
#         _z_vec = np.array(
#             [np.zeros_like(self.x), np.zeros_like(self.y), np.ones_like(self.z)]
#         )
#         N_vec = np.cross(_z_vec, h_vec, axis=0)
#         Nxy = np.sqrt(N_vec[0] * N_vec[0] + N_vec[1] * N_vec[1])
#         N_uvec = N_vec / Nxy

#         # RAAN
#         raan = np.arctan2(N_vec[1] / Nxy, N_vec[0] / Nxy)

#         # Argument of periapsis
#         sign_aop_condition = np.sum(np.cross(N_uvec, e_vec, axis=0) * h_vec, axis=0) > 0
#         sign_aop = 2 * sign_aop_condition - 1
#         aop = sign_aop * np.arccos(np.sum(e_uvec * N_uvec, axis=0))

#         # True anomaly
#         sign_ta_condition = (
#             np.sum(np.cross(e_vec, self.r_vec, axis=0) * h_vec, axis=0) > 0
#         )
#         sign_ta = 2 * sign_ta_condition - 1
#         ta = sign_ta * np.arccos(np.sum(self.r_vec * e_uvec / self.r_mag, axis=0))

#         return KeplerianState(a, e, inc, raan, aop, ta, deg=False)


# class _CartesianStateDerivative[U: (Double, Vector)](GenericState):
#     """Derivative of cartesian state

#     Derivative of a single state of a time series of states of a body in a
#     three-dimensional space in terms of cartesian coordinates. Each component
#     can be a floating point number (single state) or a numpy array of them
#     (time series of states).

#     :param dx: X component of the velocity vector [m/s]
#     :param dy: Y component of the velocity vector [m/s]
#     :param dz: Z component of the velocity vector [m/s]
#     :param ddx: X component of the acceleration vector [m/s^2]
#     :param ddy: Y component of the acceleration vector [m/s^2]
#     :param ddz: Z component of the acceleration vector [m/s^2]
#     """

#     def __init__(self, dx: U, dy: U, dz: U, ddx: U, ddy: U, ddz: U) -> None:
#         super().__init__(dx, dy, dz, ddx, ddy, ddz)
#         return None

#     @property
#     def dx(self) -> U:
#         return self.q1[0] if self.scalar else self.q1

#     @property
#     def dy(self) -> U:
#         return self.q2[0] if self.scalar else self.q2

#     @property
#     def dz(self) -> U:
#         return self.q3[0] if self.scalar else self.q3

#     @property
#     def ddx(self) -> U:
#         return self.q4[0] if self.scalar else self.q4

#     @property
#     def ddy(self) -> U:
#         return self.q5[0] if self.scalar else self.q5

#     @property
#     def ddz(self) -> U:
#         return self.q6[0] if self.scalar else self.q6

#     @property
#     def v_vec(self) -> Vector:
#         return np.array([self.dx, self.dy, self.dz])

#     @property
#     def a_vec(self) -> Vector:
#         return np.array([self.ddx, self.ddy, self.ddz])

#     @property
#     def v_mag(self) -> U:
#         return np.linalg.norm(self.v_vec, axis=0)

#     @property
#     def dv_mag(self) -> U:
#         return np.linalg.norm(self.v_vec, axis=0)

#     def times_dt(self, dt: Double) -> CartesianState:
#         """Change in cartesian state over interval dt

#         Returns the change of a cartesian state with this object as derivative
#         over a time interval dt. The main use case are numerical integrators,
#         in which the state is updated by multiplying its derivative by the
#         time step.

#         :param dt: Time step
#         :return: Change in cartesian elements over interval dt
#         """

#         return CartesianState(
#             self.dx * dt,
#             self.dy * dt,
#             self.dz * dt,
#             self.ddx * dt,
#             self.ddy * dt,
#             self.ddz * dt,
#         )

#     def add_acceleration(self, other: Self) -> None:
#         """Update the acceleration of the cartesian state derivative

#         Update the acceleration of this object by adding the components of the
#         acceleration vector of other cartesian state derivative. The use case is
#         to add up all the accelerations acting on a body during orbit propagation.
#         Since each acceleration is defined by a cartesian state derivative,
#         adding them up directly would also update the velocity, which is incorrect.

#         :param other: Cartesian state derivative with acceleration to add
#         """

#         # q4, q5 and q6 are the components of a general state vector that
#         # represent the acceleration in CartesianStateDerivative
#         self.q4 += other.q4
#         self.q5 += other.q5
#         self.q6 += other.q6

#     def __repr__(self) -> str:
#         return (
#             "Cartesian state derivative\n"
#             "---------------------------------\n"
#             f"dx: {self.dx}\n"
#             f"dy: {self.dy}\n"
#             f"dz: {self.dz}\n"
#             f"ddx: {self.ddx}\n"
#             f"ddy: {self.ddy}\n"
#             f"ddz: {self.ddz}"
#         )


# class _KeplerianState[U: (Double, Vector)](GenericState):
#     """Keplerian state vector

#     Single state of a time series of states of a body in a three-dimensional
#     space in terms of classical keplerian elements. Each component of the state
#     vector can be a floating point number (single state) or a numpy array of them
#     (time series of states).

#     Angles might be given in degrees or radians, but are always converted to
#     radians during initialization. For each angle, the method KeplerianState.angle
#     will return the value in radians, while KeplerianState.angle_deg will return
#     it in degrees.

#     Angles can be wrapped if specified by the wrap parameter, which is activated
#     by default. Wrapping will limit inclinations to the range [0, pi) and the
#     rest of the angles to [0, 2pi).

#     :param a: Semi-major axis [m]
#     :param e: Eccentricity
#     :param i: Inclination [rad]
#     :param raan: Right ascension of the ascending node [rad]
#     :param aop: Argument of periapsis [rad]
#     :param ta: True anomaly [rad]
#     :param deg: Angles are given in degrees (True) or radians (False)
#     :param wrap: Wrap angles (True) or not (False)
#     """

#     def __init__(
#         self,
#         a: U,
#         e: U,
#         i: U,
#         raan: U,
#         aop: U,
#         ta: U,
#         deg: bool = False,
#         wrap: bool = True,
#     ) -> None:
#         super().__init__(a, e, i, raan, aop, ta)

#         if deg:
#             self.q3 = np.deg2rad(self.q3)
#             self.q4 = np.deg2rad(self.q4)
#             self.q5 = np.deg2rad(self.q5)
#             self.q6 = np.deg2rad(self.q6)

#         if wrap:
#             self.q3 = self.q3 % np.pi
#             self.q4 = self.q4 % (2.0 * np.pi)
#             self.q5 = self.q5 % (2.0 * np.pi)
#             self.q6 = self.q6 % (2.0 * np.pi)

#         return None

#     @property
#     def a(self) -> U:
#         return self.q1[0] if self.scalar else self.q1

#     @property
#     def e(self) -> U:
#         return self.q2[0] if self.scalar else self.q2

#     @property
#     def i(self) -> U:
#         return self.q3[0] if self.scalar else self.q3

#     @property
#     def raan(self) -> U:
#         return self.q4[0] if self.scalar else self.q4

#     @property
#     def aop(self) -> U:
#         return self.q5[0] if self.scalar else self.q5

#     @property
#     def ta(self) -> U:
#         return self.q6[0] if self.scalar else self.q6

#     @property
#     def i_deg(self) -> U:
#         out: Any = np.rad2deg(self.i)
#         return out

#     @property
#     def raan_deg(self) -> U:
#         out: Any = np.rad2deg(self.raan)
#         return out

#     @property
#     def aop_deg(self) -> U:
#         out: Any = np.rad2deg(self.aop)
#         return out

#     @property
#     def ta_deg(self) -> U:
#         out: Any = np.rad2deg(self.ta)
#         return out

#     @property
#     def E(self) -> U:
#         """Eccentric anomaly"""
#         out: Any = 2.0 * np.arctan2(
#             np.sqrt(1.0 - self.e) * np.sin(0.5 * self.ta),
#             np.sqrt(1.0 + self.e) * np.cos(0.5 * self.ta),
#         )
#         return out

#     @property
#     def M(self) -> U:
#         """Mean anomaly"""
#         out: Any = self.E - self.e * np.sin(self.E)
#         return out

#     def period(self, mu: Double) -> U:
#         """Orbital period"""
#         out: Any = 2.0 * np.pi * np.sqrt(self.a * self.a * self.a / mu)
#         return out

#     def __sub__(self, other: Self | Double | int) -> Self:
#         print(
#             "WARNING: Keplerian state subtraction is not properly tested\n"
#             "It might produce unexpected results"
#         )

#         if not isinstance(other, KeplerianState):
#             return NotImplemented

#         di = np.unwrap(self.q3, period=180.0) - np.unwrap(other.q3, period=180.0)
#         draan = np.unwrap(self.q4, period=360.0) - np.unwrap(other.q4, period=360.0)
#         daop = np.unwrap(self.q5, period=360.0) - np.unwrap(other.q5, period=360.0)
#         dta = np.unwrap(self.q6, period=360.0) - np.unwrap(other.q6, period=360.0)

#         return type(self)(
#             self.q1 - other.q1, self.q2 - other.q2, di, draan, daop, dta, wrap=False
#         )

#     def __repr__(self) -> str:
#         return (
#             "Keplerian state vector\n"
#             "----------------------\n"
#             f"a: {self.a}\n"
#             f"e: {self.e}\n"
#             f"i: {self.i}\n"
#             f"RAAN: {self.raan}\n"
#             f"AoP: {self.aop}\n"
#             f"TA: {self.ta}"
#         )

#     def as_cartesian(self, mu: Double) -> "CartesianState":
#         """Conversion to cartesian state vector

#         Algorithm: Practical astrodynamics slides

#         :param mu: Gravitational parameter of central body [m^3/s^2]
#         """

#         # Auxiliary trigonometric relations
#         cos_Omega = np.cos(self.raan)
#         sin_Omega = np.sin(self.raan)
#         cos_omega = np.cos(self.aop)
#         sin_omega = np.sin(self.aop)
#         cos_i = np.cos(self.i)
#         sin_i = np.sin(self.i)
#         cos_theta = np.cos(self.ta)
#         sin_theta = np.sin(self.ta)

#         l1 = cos_Omega * cos_omega - sin_Omega * sin_omega * cos_i
#         l2 = -cos_Omega * sin_omega - sin_Omega * cos_omega * cos_i
#         m1 = sin_Omega * cos_omega + cos_Omega * sin_omega * cos_i
#         m2 = -sin_Omega * sin_omega + cos_Omega * cos_omega * cos_i
#         n1 = sin_omega * sin_i
#         n2 = cos_omega * sin_i

#         # Orbital radius
#         r = self.a * (1.0 - self.e * self.e) / (1.0 + self.e * cos_theta)

#         # Position in orbital plane
#         xi = r * cos_theta
#         eta = r * sin_theta

#         # Position in 3D space
#         x = l1 * xi + l2 * eta
#         y = m1 * xi + m2 * eta
#         z = n1 * xi + n2 * eta

#         # Angular momentum
#         H = np.sqrt(mu * self.a * (1.0 - self.e * self.e))

#         # Velocity
#         common = mu / H
#         e_cos_theta = self.e + cos_theta

#         dx = common * (l2 * e_cos_theta - l1 * sin_theta)
#         dy = common * (m2 * e_cos_theta - m1 * sin_theta)
#         dz = common * (n2 * e_cos_theta - n1 * sin_theta)

#         return CartesianState(x, y, z, dx, dy, dz)


# class _KeplerianStateDerivative[U: (Double, Vector)](GenericState):
#     """Derivative of keplerian state

#     Derivative of a single state of a time series of states of a body in a
#     three-dimensional space in terms of classical keplerian elements. Each
#     component can be a floating point number (single state) or a numpy array
#     of them (time series of states).

#     Angular velocities might be given in degrees or radians per second, but are
#     always converted to radians per second during initialization. For each angular
#     velocity, the method KeplerianStateDerivative.angvel will return the value in
#     radians per second, while KeplerianStateDerivative.angvel_deg will return it
#     in degrees per second.

#     :param da: Rate of change of semi-major axis [m/s]
#     :param de: Rate of change of eccentricity [1/s]
#     :param di: Rate of change of inclination [rad/s]
#     :param draan: Rate of change of right ascension of the ascending node [rad/s]
#     :param daop: Rate of change of argument of periapsis [rad/s]
#     :param dta: Rate of change of true anomaly [rad/s]
#     :param deg: Angular velocities are given in degrees per second (True) or
#         radians  per second (False)
#     """

#     def __init__(
#         self,
#         da: U,
#         de: U,
#         di: U,
#         draan: U,
#         daop: U,
#         dta: U,
#         deg: bool = True,
#     ) -> None:
#         super().__init__(da, de, di, draan, daop, dta)

#         if not deg:
#             self.q3 = np.rad2deg(self.q3)
#             self.q4 = np.rad2deg(self.q4)
#             self.q5 = np.rad2deg(self.q5)
#             self.q6 = np.rad2deg(self.q6)

#         return None

#     @property
#     def da(self) -> U:
#         return self.q1[0] if self.scalar else self.q1

#     @property
#     def de(self) -> U:
#         return self.q2[0] if self.scalar else self.q2

#     @property
#     def di(self) -> U:
#         return self.q3[0] if self.scalar else self.q3

#     @property
#     def draan(self) -> U:
#         return self.q4[0] if self.scalar else self.q4

#     @property
#     def daop(self) -> U:
#         return self.q5[0] if self.scalar else self.q5

#     @property
#     def dta(self) -> U:
#         return self.q6[0] if self.scalar else self.q6

#     def times_dt(self, dt: Double) -> KeplerianState:
#         """Change in keplerian state over interval dt

#         Returns the change of a keplerian state with this object as derivative
#         over a time interval dt. The main use case are numerical integrators,
#         in which the state is updated by multiplying its derivative by the
#         time step.

#         :param dt: Time step
#         :return: Change in keplerian elements over interval dt
#         """

#         return KeplerianState(
#             self.da * dt,
#             self.de * dt,
#             self.di * dt,
#             self.draan * dt,
#             self.daop * dt,
#             self.dta * dt,
#         )

#     def __repr__(self) -> str:
#         return (
#             "Keplerian state derivative\n"
#             "---------------------------------\n"
#             f"da: {self.da}\n"
#             f"de: {self.de}\n"
#             f"di: {self.di}\n"
#             f"dRAAN: {self.draan}\n"
#             f"dAoP: {self.daop}\n"
#             f"dTA: {self.dta}"
#         )


# # Convencience interfaces


# class _CartesianPosition[U: (Double, Vector)](GenericState):
#     """Three dimensional position vector in cartesian coordinates

#     Components might be floating point numbers (single vector) or numpy arrays
#     of them (time series of vectors).

#     :param x: X coordinate [m]
#     :param y: Y coordinate [m]
#     :param z: Z coordinate [m]
#     """

#     def __init__(self, x: U, y: U, z: U) -> None:

#         if isinstance(x, Double):
#             null = 0.0
#         else:
#             null = np.zeros_like(x)

#         super().__init__(x, y, z, null, null, null)
#         return None

#     @property
#     def x(self) -> U:
#         return self.q1[0] if self.scalar else self.q1

#     @property
#     def y(self) -> U:
#         return self.q2[0] if self.scalar else self.q2

#     @property
#     def z(self) -> U:
#         return self.q3[0] if self.scalar else self.q3

#     def asarray(self) -> Vector:
#         return np.array([self.x, self.y, self.z])

#     @property
#     def mag(self) -> U:
#         return np.linalg.norm(self.asarray(), axis=0)

#     def __repr__(self) -> str:
#         return (
#             "Position vector: \n"
#             "-----------------\n"
#             f"x: {self.x}\n"
#             f"y: {self.y}\n"
#             f"z: {self.z}\n"
#         )


# class _CartesianVelocity[U: (Double, Vector)](GenericState):
#     """Three dimensional velocity vector in cartesian coordinates

#     Components might be floating point numbers (single vector) or numpy arrays
#     of them (time series of vectors).

#     :param dx: X component of the velocity vector [m/s]
#     :param dy: Y component of the velocity vector [m/s]
#     :param dz: Z component of the velocity vector [m/s]
#     """

#     def __init__(self, dx: U, dy: U, dz: U) -> None:

#         if isinstance(dx, Double):
#             null = 0.0
#         else:
#             null = np.zeros_like(dx)

#         super().__init__(null, null, null, dx, dy, dz)
#         return None

#     @property
#     def dx(self) -> U:
#         return self.q4[0] if self.scalar else self.q4

#     @property
#     def dy(self) -> U:
#         return self.q5[0] if self.scalar else self.q5

#     @property
#     def dz(self) -> U:
#         return self.q6[0] if self.scalar else self.q6

#     def asarray(self) -> Vector:
#         return np.array([self.dx, self.dy, self.dz])

#     @property
#     def mag(self) -> U:
#         return np.linalg.norm(self.asarray(), axis=0)

#     def __repr__(self) -> str:
#         return (
#             "Velocity vector: \n"
#             "-----------------\n"
#             f"dx: {self.dx}\n"
#             f"dy: {self.dy}\n"
#             f"dz: {self.dz}\n"
#         )


# class _SphericalPosition[U: (Double, Vector)](GenericState):
#     """Three dimensional position vector in spherical coordinates

#     Components might be floating point numbers (single vector) or numpy arrays
#     of them (time series of vectors).

#     Angles can be passed in radians (default) or degrees (deg=True), but they
#     are always converted to radians during initialization. For each angle, the
#     method SphericalPosition.angle will return the value in radians, while
#     SphericalPosition.angle_deg will return it in degrees.

#     :param r: Radial distance [m]
#     :param theta: Horizontal angular distance [rad / deg]
#     :param phi: Vertical angular distance [rad / deg]
#     :param deg: Angles are given in degrees (True) or radians (False)
#     :param wrap: Wrap angles (True) or not (False)
#     """

#     def __init__(
#         self, r: U, theta: U, phi: U, deg: bool = False, wrap: bool = False
#     ) -> None:

#         if isinstance(r, Double):
#             null = 0.0
#         else:
#             null = np.zeros_like(r)

#         super().__init__(r, theta, phi, null, null, null)

#         if deg:
#             self.q2 = np.deg2rad(self.q2)
#             self.q3 = np.deg2rad(self.q3)
#         if wrap:
#             raise NotImplementedError(
#                 "Angle wrapping is not implemented for spherical coordinates"
#             )

#         return None

#     @property
#     def r(self) -> U:
#         return self.q1[0] if self.scalar else self.q1

#     @property
#     def theta(self) -> U:
#         return self.q2[0] if self.scalar else self.q2

#     @property
#     def phi(self) -> U:
#         return self.q3[0] if self.scalar else self.q3

#     @property
#     def theta_deg(self) -> U:
#         out: Any = np.rad2deg(self.theta)
#         return out

#     @property
#     def phi_deg(self) -> U:
#         out: Any = np.rad2deg(self.phi)
#         return out

#     def height(self, reference: Double) -> U:
#         return self.r - reference

#     def asarray(self) -> Vector:
#         return np.array([self.r, self.theta, self.phi])

#     def __repr__(self) -> str:
#         return (
#             "Spherical position vector\n"
#             "------------------------\n"
#             f"r: {self.r}\n"
#             f"theta: {self.theta}\n"
#             f"phi: {self.phi}\n"
#         )



from ..types.core import Vector
from ..types.time import UTC
from ..constants import arcsec
import numpy as np
from pathlib import Path


class EOP:
    """Earth Orientation Parameters

    Attributes
    ----------
    xp, yp : float
        Celestial pole coordinates
    ut1_utc : float
        Difference between UT1 and UTC [includes leap seconds]
    lod : float
        Length of day
    ddpsi, ddeps : float
        Nutation corrections
    dx, dy : float
        Celestial pole offsets

    Sources
    -------
    https://ggos.org/item/earth-orientation-parameter
    """

    def __init__(self, eop_data: Vector, tai_utc: int | None = None) -> None:

        if tai_utc is None:
            tai_utc = int(eop_data[-1])

        self.xp = eop_data[0] * arcsec
        self.yp = eop_data[1] * arcsec
        self.ut1_utc = eop_data[2] + tai_utc
        self.lod = eop_data[3]
        self.ddpsi = eop_data[4] * arcsec
        self.ddeps = eop_data[5] * arcsec
        self.dx = eop_data[6] * arcsec
        self.dy = eop_data[7] * arcsec

        return None

    @classmethod
    def at_epoch(cls, utc: UTC) -> "EOP":
        """Earth Orientation Parameters at epoch

        Retrieve EOP data at given epoch by computing a linear interpolation of
        parameters from the two closest times in the EOP data file.

        :param utc: UTC date and time
        """

        # Approximate the MJD of the input date
        epoch = utc.as_jd()
        mjd_int = int(epoch.mjd)

        # Find closest epochs in EOP data
        eop_data = np.load(Path(__file__).parent / "eop.npy").T
        idx_lower = np.nonzero(eop_data[3] >= mjd_int)[0][0]
        idx_upper = idx_lower + 1
        lower = eop_data[:, idx_lower]
        upper = eop_data[:, idx_upper]

        # Adjust UT1-UTC column in case leap second occurs between lines
        tai_utc = int(lower[-1])
        lower[3] -= lower[9]
        upper[3] -= upper[9]

        # Linear interpolation
        dt = epoch.mjd - lower[0]
        delta_eop = upper - lower
        interp_eop = lower[1:] + dt * delta_eop[1:] / delta_eop[0]

        return cls(interp_eop, tai_utc)



from ..types import Vector, Double


class RadarObservation[T: (Double, Vector)]:

    def __init__(
        self, epoch: T, range: Double, azimuth: Double, elevation: Double
    ) -> None:
        self.epoch = epoch
        self.range = range
        self.azimuth = azimuth
        self.elevation = elevation

        return None



from .base import Force
from .. import types as nt


class PointMass(Force):

    def __init__(self, mu: nt.Double) -> None:

        self.mu = mu

        return None

    def __call__(
        self, t: nt.Double, s: nt.CartesianState, fr: nt.Double = 0.0
    ) -> nt.CartesianStateDerivative:

        f = -self.mu / (s.r_mag * s.r_mag * s.r_mag)

        return nt.CartesianStateDerivative(s.dx, s.dy, s.dz, f * s.x, f * s.y, f * s.z)



from . import core as nt
from datetime import datetime
import numpy as np
from typing import Any, Iterator, Sequence, Self, Union, Literal
from pathlib import Path
from ..constants import day
import traceback

# from ..data.formats import EOP


class JulianDay[T: (nt.Double, nt.Vector)]:
    """Julian day

    .. warning:: You might pass the integral and fractional parts of the Julian day as a single number, but this will result in a loss of precision due to floating point arithmetic.

    A Julian Day is specified by an integral part (day), which corresponds to
    12:00:00 of a certain day, and a fractional part (time), which represents the
    amount of time that has passed since that moment as a fraction of a day.

    NOTE: Accurate to 16 decimal places if integral part ends in 0.5 and to 15 - int(log_10(jd_int)) decimal places otherwise. For example, with (2451545.5, None) as input, the fractional part is accurate to 16 decimal places, but with (2451545.76, None) as input, the fractional part is accurate to 10 decimal places.
    """

    def __init__(
        self,
        jd_int: T,
        jd_frac: T | None = None,
        ref: Literal["J2000", "MJD"] | None = None,
    ) -> None:

        # Get type
        if nt.is_double(jd_int):
            if jd_frac is not None and not nt.is_double(jd_frac):
                raise ValueError("Failed to initialize JulianDay: Invalid input")
            self.scalar = True
        elif isinstance(jd_int, np.ndarray):
            if isinstance(jd_frac, np.ndarray) and jd_frac.dtype == np.dtype("O"):
                raise ValueError("Failed to initialize JulianDay: Invalid input")
            self.scalar = False
        else:
            raise ValueError("Failed to initialize JulianDay: Invalid input type")

        # Generate array containers
        q1: Any = np.array([jd_int], dtype=np.float64).ravel()
        if jd_frac is None:
            q2: Any = np.zeros_like(q1)
        else:
            q2: Any = np.array([jd_frac], dtype=np.float64).ravel()

        if q1.size != q2.size:
            raise ValueError("Integral and fractional parts must have the same size")

        # Add reference epoch if necessary
        if ref is not None:

            if ref == "J2000":
                q1 += 2451545.0
            elif ref == "MJD":
                q1 += 2400000.0
                q2 += 0.5
            else:
                raise ValueError(
                    "Failed to initialize JulianDay: Invalid reference epoch"
                )

        self.q1, self.q2 = self.__split(q1, q2)

        return None

    def __split(self, jd_int: Any, jd_frac: Any) -> tuple[Any, Any]:
        """Split Julian date into integral and fractional parts"""

        day = np.round(jd_int) - 0.5
        time = jd_int - day
        dI, dF = np.divmod(jd_frac, 1.0)
        ddI, dF = np.divmod(time + dF, 1.0)
        return day + dI + ddI, dF

    @property
    def size(self) -> int:
        return self.q1.size

    @property
    def day(self) -> T:
        """Integral part of the Julian Day"""
        return self.q1[0] if self.scalar else self.q1

    @property
    def time(self) -> T:
        """Fractional part of the Julian Day"""
        return self.q2[0] if self.scalar else self.q2

    @property
    def jd(self) -> T:
        """Julian date and time"""
        return self.day + self.time

    @property
    def mday(self) -> T:
        """Integral part of modified Julian Day"""
        return (self - 2400000.5).day

    @property
    def mtime(self) -> T:
        """Fractional part of modified Julian Day"""
        return (self - 2400000.5).time

    @property
    def mjd(self) -> T:
        """Modified Julian date and time"""
        return (self - 2400000.5).jd

    @property
    def dt(self) -> T:
        """Days past initial epoch"""
        return self.jd - self[0].jd

    def __repr__(self) -> str:

        out = "Julian Day\n"
        out += "-" * len(out) + "\n"
        out += f"Day: {self.day}\n"
        out += f"Time: {self.time}\n"
        return out

    def __sub__(self, other: object) -> Union["JulianDay[T]", "DeltaJD[T]"]:

        if nt.is_double(other):
            dI, dF = np.divmod(other, 1.0)
            return JulianDay(self.day - dI, self.time - dF)
        elif isinstance(other, np.ndarray) and other.dtype != np.dtype("O"):
            if other.size != self.size:
                raise ValueError("Failed to perform subtraction: Sizes do not match")
            dI, dF = np.divmod(other, 1.0)
            day: Any = self.day - dI
            time: Any = self.time - dF
            return JulianDay(day, time)
        elif isinstance(other, self.__class__):
            if self.size != other.size:
                raise ValueError("Failed to perform subtraction: Sizes do not match")

            delta_day = self.day - other.day
            delta_time = self.time - other.time
            return DeltaJD(delta_day, delta_time)
        else:
            raise TypeError("Failed to perform subtraction. Invalid type")

    def __add__(self, other: object) -> "JulianDay[T]":

        if nt.is_double(other):
            dI, dF = np.divmod(other, 1.0)
            return JulianDay(self.day + dI, self.time + dF)
        elif isinstance(other, np.ndarray) and other.dtype != np.dtype("O"):
            if other.size != self.size:
                raise ValueError("Failed to perform addition: Sizes do not match")
            dI, dF = np.divmod(other, 1.0)
            day: Any = self.day + dI
            time: Any = self.time + dF
            return JulianDay(day, time)
        else:
            raise TypeError("Failed to perform addition. Invalid type")

    def __getitem__(self, index: int | slice) -> "JulianDay[T]":
        return JulianDay(self.q1[index], self.q2[index])

    def __iter__(self) -> Iterator["JulianDay[T]"]:
        return iter([self.__getitem__(idx) for idx in range(self.size)])

    def __eq__(self, other: object) -> bool:
        assert isinstance(other, JulianDay)
        same_day = np.allclose(self.day, other.day, atol=1e-15, rtol=0.0)
        same_time = np.allclose(self.time, other.time, atol=1e-15, rtol=0.0)
        return same_day and same_time

    def split(self, gap: nt.Double) -> tuple[Sequence[int], Sequence[Self]]:
        """Split time series of Julian dates into intervals

        :param gap: Maximum gap between consecutive epochs in the same interval
        :return: Tuple containing indices of split points and sequence of intervals
        """

        delta = self[1:] - self[:-1]
        idx_list = np.argwhere(delta.jd > gap).ravel() + 1
        passes = []
        low = 0
        for idx in idx_list:
            passes.append(self[low:idx])
            low = idx
        passes.append(self[low:])
        idx_list = [0] + idx_list.tolist()
        return idx_list, passes

    def as_calendar(self) -> "CalendarDate" | Sequence["CalendarDate"]:
        """Conversion to calendar date and time

        Algorithm: Practical astrodynamics slides - 2.17
        """

        jd0 = self.day + 0.5
        L1 = np.trunc(jd0 + 68569)
        L2 = np.trunc(4 * L1 / 146097)
        L3 = L1 - np.trunc((146097 * L2 + 3) / 4)
        L4 = np.trunc(4000 * (L3 + 1) / 1461001)
        L5 = L3 - np.trunc(1461 * L4 / 4) + 31
        L6 = np.trunc(80 * L5 / 2447)
        L7 = np.trunc(L6 / 11)
        day = L5 - np.trunc(2447 * L6 / 80)
        month = L6 + 2 - 12 * L7
        year = 100 * (L2 - 49) + L4 + L7
        hour = np.trunc(self.time * 24)
        rem = self.time * 24 - hour
        minute = np.trunc(rem * 60)
        rem = rem * 60 - minute
        second = np.trunc(rem * 60)
        micro_second = np.trunc((rem * 60 - second) * 1e6)

        if nt.is_double(self.day):
            return CalendarDate(
                int(year),
                int(month),
                int(day),
                int(hour),
                int(minute),
                int(second),
                int(micro_second),
            )

        elif isinstance(self.day, np.ndarray):
            return [
                CalendarDate(
                    int(year[i]),
                    int(month[i]),
                    int(day[i]),
                    int(hour[i]),
                    int(minute[i]),
                    int(second[i]),
                    int(micro_second[i]),
                )
                for i in range(self.day.size)
            ]
        else:
            raise TypeError("Failed to convert to calendar date: Invalid type")

    def save(self, path: str | Path) -> None:
        """Save Julian date to Numpy binary file

        :param path: Path to file to save
        """
        if isinstance(path, str):
            path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)
        np.save(path, np.array([self.day, self.time]))
        return None

    @classmethod
    def load(
        cls,
        path: str | Path,
        ref: Literal["J2000", "MJD"] | None = None,
        unit: Literal["s", "d"] = "d",
        relative: bool = True,
    ) -> Self:
        """Load Julian date from Numpy binary file

        :param path: Path to file to load
        """

        # If extension is missing, assume .npy
        path = Path(path)
        if not path.suffix:
            path = path.with_suffix(".npy")

        # Generate absolute path
        cwd = Path("/")
        if relative:
            cwd = Path(traceback.extract_stack()[-2].filename).parent
        path = cwd / path

        # Load output
        match path.suffix:
            case ".npy":
                return cls(*np.load(path), ref=ref)
            case ".dat":
                data = np.loadtxt(path).T
                if unit == "s":
                    data /= day
                if data.shape[0] == 1 or data.shape[0] == 7:
                    return cls(data[0], ref=ref)
                elif data.shape[0] == 2 or data.shape[0] == 8:
                    return cls(data[0], data[1], ref=ref)
                else:
                    raise ValueError("Failed to load Julian date: Invalid data format")
            case _:
                raise ValueError("Failed to load Julian date: Invalid file extension")

    @classmethod
    def from_tudat(
        cls,
        state_history: dict[nt.Double, list[nt.Double]],
        ref: Literal["J2000", "MJD"] | None = None,
        unit: Literal["s", "d"] = "d",
    ) -> Self:
        """Load Julian date from TUDAT state history

        :param state_history: Dictionary containing state history data
        """
        data = np.array(list(state_history.keys())) / day
        return cls(data, ref=ref)


class DeltaJD[T: (nt.Double, nt.Vector)]:
    """Difference between two Julian days

    NOTE: You are not supposed to instantiate DeltaJD. You can subtract Double or
    Vector types directly from a JulianDay object.

    :param dI: Integral part of the difference
    :param dF: Fractional part of the difference
    """

    def __init__(self, dI: T, dF: T) -> None:

        # Get typea
        if nt.is_double(dI):
            if dF is not None and not nt.is_double(dF):
                raise ValueError("Failed to initialize JulianDay: Invalid input")
            self.scalar = True
        elif isinstance(dI, np.ndarray):
            if not isinstance(dF, np.ndarray) or dF.dtype == np.dtype("O"):
                raise ValueError("Failed to initialize JulianDay: Invalid input")
            self.scalar = False
        else:
            raise ValueError("Failed to initialize JulianDay: Invalid input type")

        q1: Any = np.array([dI], dtype=np.float64).ravel()
        q2: Any = np.array([dF], dtype=np.float64).ravel()

        if q1.size != q2.size:
            raise ValueError("Integral and fractional parts must have the same size")

        self.q1, self.q2 = self.__split(q1, q2)

        return None

    def __split(self, dI: Any, dF: Any) -> tuple[Any, Any]:
        """Split Julian date into integral and fractional parts"""

        ddI = np.round(dI)
        ddF = dI - ddI
        deltaI, deltaF = np.divmod(dF, np.where(dF < 0, -1.0, 1.0))
        ddF = ddF + deltaF
        ddeltaI, deltaF = np.divmod(ddF, np.where(ddF < 0, -1.0, 1.0))
        return ddI + deltaI + ddeltaI, deltaF

    @property
    def size(self) -> int:
        return self.q1.size

    @property
    def day(self) -> T:
        return self.q1[0] if self.scalar else self.q1

    @property
    def time(self) -> T:
        return self.q2[0] if self.scalar else self.q2

    @property
    def jd(self) -> T:
        return self.day + self.time


class CalendarDate(datetime):
    """Base class for calendar date types"""

    def as_list(self) -> list[int]:
        return [self.year, self.month, self.day, self.hour, self.minute, self.second]

    def as_jd(self) -> "JulianDay":
        """Conversion to Julian date and time

        Algorithm: Practical astrodynamics slides - 2.16
        """

        C = np.trunc((self.month - 14) / 12)
        jd0 = self.day - 32075 + np.trunc(1461 * (self.year + 4800 + C) / 4)
        jd0 += np.trunc(367 * (self.month - 2 - C * 12) / 12)
        jd0 -= np.trunc(3 * np.trunc((self.year + 4900 + C) / 100) / 4)
        jd = jd0 - 0.5
        fr = self.hour / 24.0 + self.minute / 1440.0 + self.second / 86400.0

        return JulianDay(jd, fr)


# class JulianDate[T: (Double, Vector)]:
#     """Julian date and time"""

#     def __init__(self, jd: T, fr: T | None = None) -> None:

#         self.scalar = False
#         if isinstance(jd, Double):
#             self.scalar = True

#         self.q1: Any = np.array([jd], dtype=np.float64).ravel()
#         if fr is None:
#             self.q2: Any = np.zeros_like(self.q1)
#         else:
#             self.q2: Any = np.array([fr], dtype=np.float64).ravel()

#         return None

#     @property
#     def size(self) -> int:
#         return self.q1.size

#     @property
#     def day(self) -> T:
#         return self.q1[0] if self.scalar else self.q1

#     @property
#     def frac(self) -> T:
#         return self.q2[0] if self.scalar else self.q2

#     @property
#     def mint(self) -> T:
#         return self.day - 2400000.5

#     @property
#     def jd(self) -> T:
#         return self.day + self.frac

#     @property
#     def mjd(self) -> T:
#         return self.mint + self.frac

#     def __eq__(self, other: object) -> bool:
#         assert isinstance(other, JulianDate)
#         is_this_jd = (self.day == other.day) and (self.frac == other.frac)
#         return is_jd and is_this_jd

#     def __repr__(self) -> str:

#         limit = 5 if self.size > 5 else self.size
#         out = "Integral:\t"
#         for i in range(limit):
#             out += f"{self.q1[i]:.15e} "
#         if self.size > 5:
#             out += "..."
#         out += "\nFractional:\t"
#         for i in range(limit):
#             out += f"{self.q2[i]:.15e} " if self.q2[i] != 0.0 else "0 "
#         if self.size > 5:
#             out += "..."
#         return out

#     def __add__(self, other: Self) -> "JulianDate[T]":

#         return JulianDate(self.day + other.day, self.frac + other.frac)

#     def __sub__(self, other: Self) -> "JulianDate":

#         return JulianDate(self.day - other.day, self.frac - other.frac)

#     def __getitem__(self, index: int | slice) -> "JulianDate":

#         return JulianDate(self.q1[index], self.q2[index])

#     def __iter__(self) -> Iterator["JulianDate"]:

#         return iter([self.__getitem__(idx) for idx in range(self.size)])

#     def as_date(self) -> Date | Sequence[Date]:
#         """Conversion to calendar date and time

#         Algorithm: Practical astrodynamics slides - 2.17
#         """

#         jd0 = self.day + 0.5
#         L1 = np.trunc(jd0 + 68569)
#         L2 = np.trunc(4 * L1 / 146097)
#         L3 = L1 - np.trunc((146097 * L2 + 3) / 4)
#         L4 = np.trunc(4000 * (L3 + 1) / 1461001)
#         L5 = L3 - np.trunc(1461 * L4 / 4) + 31
#         L6 = np.trunc(80 * L5 / 2447)
#         L7 = np.trunc(L6 / 11)
#         day = L5 - np.trunc(2447 * L6 / 80)
#         month = L6 + 2 - 12 * L7
#         year = 100 * (L2 - 49) + L4 + L7
#         hour = np.trunc(self.frac * 24)
#         rem = self.frac * 24 - hour
#         minute = np.trunc(rem * 60)
#         rem = rem * 60 - minute
#         second = np.round(rem * 60)
#         micro_second = np.trunc((rem * 60 - second) * 1e6)

#         if isinstance(self.day, Double):
#             return Date(
#                 int(year),
#                 int(month),
#                 int(day),
#                 int(hour),
#                 int(minute),
#                 int(second),
#                 int(micro_second),
#             )

#         elif isinstance(self.day, np.ndarray):
#             return [
#                 Date(
#                     int(year[i]),
#                     int(month[i]),
#                     int(day[i]),
#                     int(hour[i]),
#                     int(minute[i]),
#                     int(second[i]),
#                     int(micro_second[i]),
#                 )
#                 for i in range(self.day.size)
#             ]
#         else:
#             raise TypeError("Unexpected type for Julian date")

#     def split(self, gap: Double) -> tuple[list[int], list["JulianDate"]]:
#         """Split time series of Julian dates into intervals

#         :param gap: Maximum gap between consecutive epochs
#         """
#         idx_list = np.argwhere(np.diff(self.jd) > gap).ravel() + 1
#         passes = []
#         low = 0
#         for idx in idx_list:
#             passes.append(self[low:idx])
#             low = idx
#         return idx_list.tolist(), passes

#     @classmethod
#     def load(cls, path: str | Path) -> Self:
#         """Load Julian date from Numpy binary file

#         :param path: Path to file to load
#         """
#         data = np.load(path)
#         return cls(data)


class UTC(CalendarDate):
    pass

    # def as_ut1(self) -> "UT1":

    #     offset = EOP.at_epoch(self).ut1_utc
    #     print(offset)
    #     exit(0)


class UT1(CalendarDate):
    pass



from .base import Force
from .. import types as nt
import numpy as np


class Coriolis(Force):

    def __init__(self, omega: nt.Double) -> None:
        self.omega = np.array([0.0, 0.0, omega])
        return None

    def __call__(
        self, t: nt.Double, s: nt.CartesianState, fr: nt.Double = 0.0
    ) -> nt.CartesianStateDerivative:
        return nt.CartesianStateDerivative(
            s.dx, s.dy, s.dz, *(-2.0 * np.cross(self.omega, s.v_vec, axis=0))
        )


class Centrifugal(Force):

    def __init__(self, omega: nt.Double) -> None:
        self.omega = np.array([0.0, 0.0, omega])
        return None

    def __call__(
        self, t: nt.Double, s: nt.CartesianState, fr: nt.Double = 0.0
    ) -> nt.CartesianStateDerivative:
        return nt.CartesianStateDerivative(
            s.dx,
            s.dy,
            s.dz,
            *(-np.cross(self.omega, np.cross(self.omega, s.r_vec, axis=0), axis=0)),
        )



"""
Types submodule
================

.. currentmodule:: nastro.types

Type aliases
-------------

=============== ==============================================================
`Double`        Double precision floating point number
`Vector`        One-dimensional array of double precision floating point numbers
`ArrayLike`     Sequence of elements that can be converted to a numpy array
=============== ==============================================================

State definition
-----------------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    GenericState
    CartesianState
    CartesianStateDerivative
    CartesianPosition
    CartesianVelocity
    KeplerianState
    KeplerianStateDerivative

Time
-----

.. autosummary::
    :toctree: generated/
    :nosignatures:

    JulianDay
    CalendarDate
    UTC

"""

from .core import (
    Scalar,
    Double,
    Vector,
    Array,
    is_scalar,
    is_double,
    is_vector,
    is_array,
)

from .state import (
    GenericState,
    CartesianState,
    CartesianStateDerivative,
    KeplerianState,
    KeplerianStateDerivative,
    CartesianPosition,
    CartesianVelocity,
)

from .time import CalendarDate, JulianDay, UTC

from .conversion import time_to_mean_anomaly

__all__ = [
    # Core
    "Scalar",
    "Double",
    "Vector",
    "Array",
    "is_scalar",
    "is_double",
    "is_vector",
    "is_array",
    # State
    "GenericState",
    "CartesianState",
    "CartesianStateDerivative",
    "KeplerianState",
    "KeplerianStateDerivative",
    "CartesianPosition",
    "CartesianVelocity",
    # Time
    "CalendarDate",
    "JulianDay",
    "UTC",
    "time_to_mean_anomaly",
]



from ..forces import ForceModel
from .. import types as nt, constants as nc
import numpy as np
from typing import Any
from scipy.integrate import solve_ivp


class SimplePropagator:

    def __init__(self, model: ForceModel) -> None:

        self.model = model

        return None

    def fun(self, t: nt.Double, s: nt.Vector) -> nt.Vector:
        return self.model(t, nt.CartesianState(*s)).asarray.ravel()

    def propagate_scipy(
        self,
        t0: nt.Double,
        tend: nt.Double,
        s0: nt.CartesianState,
        solver: str = "RK45",
        atol: nt.Double = 1e-10,
        rtol: nt.Double = 1e-10,
    ) -> tuple[nt.Vector, nt.CartesianState]:

        sol = solve_ivp(
            self.fun,
            (t0, tend),
            s0.asarray.ravel(),
            method=solver,
            atol=atol,
            rtol=rtol,
        )

        return sol.t, nt.CartesianState(*sol.y)


class SimpleCartesianPropagator:

    def __init__(
        self,
        model: ForceModel,
        t0: nt.Double,
        s0: nt.CartesianState,
        tend: nt.Double,
        dt: nt.Double,
    ) -> None:

        self.model = model
        self.now = t0
        self.tend = tend
        if not s0.scalar:
            raise TypeError("Initial state must be scalar")
        self.s = s0
        self.h = dt

        # Termination criteria
        EPS = 1e-14
        self.done = max(EPS, 0.1 * dt)
        self.status = "running"

        return None

    def fun(self, t: nt.Double, s: nt.Vector) -> nt.Vector:
        return self.model(t, nt.CartesianState(*s)).asarray.ravel()

    def scipy_propagate(
        self, solver: str = "RK45", atol: nt.Double = 1e-12, rtol: nt.Double = 1e-12
    ) -> tuple[nt.Vector, nt.CartesianState]:

        sol = solve_ivp(
            self.fun,
            (self.now, self.tend),
            self.s.asarray.ravel(),
            method=solver,
            atol=atol,
            rtol=rtol,
        )

        return sol.t, nt.CartesianState(*sol.y)

    def propagate(self) -> tuple[nt.Vector, nt.CartesianState]:

        out_state = nt.CartesianState(*self.s.asarray)
        out_time = [self.now]

        # Main loop
        status = None
        while status is None:

            # Perform integration step
            self.step()

            # Check status after step
            if self.status == "finished":
                status = 0
            elif self.status == "failed":
                status = 1
                break

            # Update output
            out_time.append(self.now)
            out_state.append(self.s)

        return np.array(out_time), out_state

    def step(self) -> None:

        # Check status before performing step
        if self.status != "running":
            raise RuntimeError("Attempted step on non-running propagator")

        # Perform step
        success = self._step_impl()

        # Check for termination condition
        if not success:
            self.status = "failed"
        else:
            if self.tend - self.now < self.done:
                self.status = "finished"

        return None

    def _step_impl(self) -> bool:
        raise NotImplementedError("Subclasses must implement this method")


class CartesianPropagator[U: (nt.Double, nt.JulianDay)]:
    """Base class for cartesian propagators

    Parameters
    -----------
    model : ForceModel
        Dynamical model representing the accelerations acting on the body
    t0, tend : Double | JulianDay[Double]
        Initial and final epochs
    s0 : CartesianState[Double]
        Initial state
    dt : Double
        Time step. Assumed to be in seconds if t0 and tend are Doubles and days
        if they are JulianDays
    """

    def __init__(
        self,
        model: ForceModel,
        t0: U,
        s0: nt.CartesianState[nt.Double],
        tend: U,
        dt: nt.Double,
    ) -> None:

        self.model = model

        # Initial and final epoch and step
        if isinstance(t0, nt.JulianDay) and isinstance(tend, nt.JulianDay):
            self.is_day = True
            self.now_int, self.now_fr = t0.day, t0.time
            self.tend_int, self.tend_fr = tend.day, tend.time
            dt /= nc.day
        elif nt.is_double(t0) and nt.is_double(tend):
            self.is_day = False
            self.now_int, self.now_fr = t0, 0.0
            self.tend_int, self.tend_fr = tend, 0.0
        else:
            raise TypeError("Unexpected type for initial or final epoch")

        self.now = self.now_int + self.now_fr
        self.tend = self.tend_int + self.tend_fr

        # Initial state
        if not s0.scalar:
            raise TypeError("Initial state must be scalar")
        self.s = s0

        # Step size
        delta_int = self.tend_int - self.now_int
        delta_fr = self.tend_fr - self.now_fr
        steps = (delta_int // dt) + (delta_fr // dt)
        self.h = (delta_int / steps) + (delta_fr / steps)
        self.h_int = np.trunc(self.h)
        self.h_fr = self.h - self.h_int

        if self.is_day:
            self.h_sec = self.h_int * nc.day + self.h_fr * nc.day
        else:
            self.h_sec = self.h_int + self.h_fr

        # Termination criteria
        self.EPS = 1e-14
        self.done_int = max(self.EPS, 0.1 * self.h_int)
        self.done_fr = max(self.EPS, 0.1 * self.h_fr)

        # Status
        self.status = "running"

        return None

    def _update_epoch(self) -> tuple[nt.Double, nt.Double]:

        new_fr = self.now_fr + self.h_fr
        int_part = np.trunc(new_fr)
        new_fr -= int_part
        new_int = self.now_int + self.h_int + int_part

        return new_int, new_fr

    def _get_steps(self) -> tuple[nt.Double, nt.Double]:

        if self.is_day:
            return self.h_int * nc.day, self.h_fr * nc.day
        else:
            return self.h_int, self.h_fr

    def propagate(self, frac: bool = True) -> tuple[U, nt.CartesianState]:
        """Propagate initial state"""

        out_sol = nt.CartesianState(*self.s.asarray)
        out_time = [self.now_int]
        out_time_fr = [self.now_fr]

        # Main loop
        prop_status = None
        while prop_status is None:

            # Perform integration step
            self.step()

            # Check status after step
            if self.status == "finished":
                prop_status = 0
            elif self.status == "failed":
                prop_status = 1
                break

            # Update state and time
            out_sol.append(self.s)
            out_time.append(self.now_int)
            out_time_fr.append(self.now_fr)

        # Generate output
        time_int = np.array(out_time, dtype=np.float64)
        time_fr = np.array(out_time_fr, dtype=np.float64)
        time: Any = (
            nt.JulianDay(time_int, time_fr) if self.is_day else time_int + time_fr
        )

        return time, out_sol

    def step(self) -> None:
        """Perform a single integration step"""

        # Check status before performing step
        if self.status != "running":
            raise RuntimeError("Attempted step on non-running propagator")

        # Perform step
        success = self._step_impl()

        # Check for termination condition
        if not success:
            self.status = "failed"
        else:
            if (self.tend_int - self.now_int < self.done_int) and (
                self.tend_fr - self.now_fr < self.done_fr
            ):
                self.status = "finished"

        return None

    def _step_impl(self) -> bool:
        raise NotImplementedError("Subclasses must implement this method")



from ..types.time import JulianDay, CalendarDate
from ..types.core import Vector
import numpy as np
from pathlib import Path
from ..constants import arcsec
from typing import TypeVar
from datetime import datetime


THIS_FILE = Path(__file__)


class EOP:
    """Earth Orientation Parameters

    Sources
    -------
    https://ggos.org/item/earth-orientation-parameter
    """

    def __init__(self, eop_data: Vector, tai_utc: int | None = None) -> None:

        if tai_utc is None:
            tai_utc = int(eop_data[-1])

        self.xp = eop_data[0] * arcsec
        self.yp = eop_data[1] * arcsec
        self.ut1_utc = eop_data[2] + tai_utc
        self.lod = eop_data[3]
        self.ddpsi = eop_data[4] * arcsec
        self.ddeps = eop_data[5] * arcsec
        self.dx = eop_data[6] * arcsec
        self.dy = eop_data[7] * arcsec

        return None

    @classmethod
    def at_epoch(cls, utc: CalendarDate) -> "EOP":
        """Earth Orientation Parameters at epoch

        Retrieve EOP data at given epoch by computing a linear interpolation of
        parameters from the two closest times in the EOP data file.

        :param utc: UTC date and time
        """

        # Approximate the MJD of the input date
        epoch = utc.as_jd()
        mjd_int = int(epoch.mjd)

        # Find closest epochs in EOP data
        eop_data = np.load(THIS_FILE.parent / "sources/eop.npy").T
        idx_lower = np.nonzero(eop_data[3] >= mjd_int)[0][0]
        idx_upper = idx_lower + 1
        lower = eop_data[:, idx_lower]
        upper = eop_data[:, idx_upper]

        # Adjust UT1-UTC column in case leap second occurs between lines
        tai_utc = int(lower[-1])
        lower[3] -= lower[9]
        upper[3] -= upper[9]

        # Linear interpolation
        dt = epoch.mjd - lower[0]
        delta_eop = upper - lower
        interp_eop = lower[1:] + dt * delta_eop[1:] / delta_eop[0]

        # Convert output to arcseconds
        return EOP(interp_eop, tai_utc)


class XYS:

    def __init__(self, xys_data: Vector) -> None:

        self.year = xys_data[0]
        self.month = xys_data[1]
        self.day = xys_data[2]
        self.mjd = xys_data[3]
        self.x = xys_data[4]
        self.y = xys_data[5]
        self.s = xys_data[6]

        return None

    @classmethod
    def interpolate(cls, x: Vector, y: Vector, xx: float, order: int = 11):

        points = order + 1

        if len(x) < points:
            raise ValueError("Not enough data points for interpolation")

        # Compute number of elements on either side of middle element to grab
        nn, rem = np.divmod(points, 2)

        # Find index such that x[row0] < xx < x[row0 + 1]
        row0 = np.nonzero(x < xx)[0][-1]

        # Trim data set
        if rem == 0:
            # Adjust row0 in case near data set endpoints
            if (points == len(x)) or (row0 < nn - 1):
                row0 = nn - 1
            elif row0 > (len(x) - nn):
                row0 = len(x) - nn - 1

            # Trim to relevant data points
            x_trimed = x[row0 - nn + 1 : row0 + nn + 1]
            y_trimed = y[:, row0 - nn + 1 : row0 + nn + 1]
        else:
            # Adjust row0 in case near data set endpoints
            if (points == len(x)) or (row0 < nn):
                row0 = nn
            elif row0 > len(x) - nn:
                row0 = len(x) - nn - 1
            else:
                if (xx - x[row0] > 0.5) and (row0 + 1 + nn < len(x)):
                    row0 += 1

            # Trim to relevant data points
            x_trimed = x[row0 - nn : row0 + nn + 1]
            y_trimed = y[:, row0 - nn : row0 + nn + 1]

        # Compute coefficients (Didn't vectorize due to floating point)
        Pj = np.ones((1, points))

        for jj in range(points):
            for ii in range(points):
                if jj != ii:
                    Pj[0, jj] = (
                        Pj[0, jj] * (x_trimed[ii] - xx) / (x_trimed[ii] - x_trimed[jj])
                    )

        return np.dot(Pj, y_trimed.T)[0]

    @classmethod
    def at_epoch(cls, tt: JulianDay):

        xys_data = np.load(THIS_FILE.parent / "sources/xys.npy")

        # Compute MJD and round to nearest integer
        mjd_int = int(tt.mjd)

        # Number of additional data points to include on either side
        n = 10

        # Find closest epochs in XYS data
        mjd_data = xys_data[3]
        if mjd_int not in mjd_data:
            raise ValueError("Requested epoch not in XYS time range")

        mjd_idx = np.nonzero(mjd_data == mjd_int)[0][0]
        low = 0 if mjd_int <= mjd_data[0] + n else mjd_idx - n
        high = -1 if mjd_int > mjd_data[0] - n else mjd_idx + n
        xys_data = xys_data[:, low:] if high == -1 else xys_data[:, low:high]

        x_data = xys_data[3]
        y_data = xys_data[4:]
        x, y, s = cls.interpolate(x_data, y_data, tt.mjd, 11) * arcsec
        return x, y, s



import numpy as np
import numpy.typing as npt
from typing import Sequence, TypeVar, Any, TypeGuard

Scalar = int | float | np.floating | np.integer
Double = float | np.float64
Vector = npt.NDArray[np.float64]
Array = npt.NDArray[Any] | Sequence[int | float]
T = TypeVar("T", Double, Vector)


def is_scalar(x: Any) -> TypeGuard[Scalar]:
    return isinstance(x, int) or isinstance(x, float) or isinstance(x, np.number)


def is_double(x: Any) -> TypeGuard[Double]:
    return isinstance(x, float)


def is_vector(x: Any) -> TypeGuard[Vector]:
    return isinstance(x, np.ndarray) and is_double(x[0])


def is_array(x: Any) -> TypeGuard[Array]:
    return (isinstance(x, np.ndarray) or isinstance(x, Sequence)) and is_scalar(x[0])



from ..types import Vector, Double, CartesianPosition


class Limits:

    min: Double
    max: Double


class Radar:

    def __init__(
        self,
        coordinates: CartesianPosition,
        range_range: tuple[Double, Double],
        elevation_range: tuple[Double, Double],
        azimuth_range: tuple[Double, Double],
        horizontal_FOV: tuple[Double, Double],
        vertical_FOV: tuple[Double, Double],
    ) -> None:

        self.coordinates = coordinates
        self.rho_lims = Limits(*range_range)
        self.el_lims = Limits(*elevation_range)
        self.az_lims = Limits(*azimuth_range)
        self.hfov_lims = Limits(*horizontal_FOV)
        self.vfov_lims = Limits(*vertical_FOV)

        return None



import numpy as np
from typing import Self, Any, Iterator
from pathlib import Path
from .core import Double, Vector
from ..frames.frames import Frame, ICRF


# JUST HERE AS REFERENCE FOR FUTURE FRAME IMPLEMENTATION


class GenericState[U: (Double, Vector)]:
    """Base class for state vectors

    Implements operations and basic functionality that is common to all types
    of state vectors and implements the containers that are used internally
    to manipulate their components. The elements q1 to q6 are representations
    of the six components of a state vector as numpy arrays.

    This class is not meant to be instantiated directly, but to be subclassed
    to create specific types of state vectors: cartesian, keplerian, etc.
    """

    def __init__(
        self,
        q1: U,
        q2: U,
        q3: U,
        q4: U,
        q5: U,
        q6: U,
        deg: bool = False,
        wrap: bool = True,
        frame: type[Frame] = ICRF,
        _components: dict[str, str] | None = None,
        _angles: dict[str, tuple[float, float]] | None = None,
    ) -> None:

        # Specify components and angles
        if _components is not None:
            self.components = _components
        else:
            self.components = {
                "q1": "q1",
                "q2": "q2",
                "q3": "q3",
                "q4": "q4",
                "q5": "q5",
                "q6": "q6",
            }
        if _angles is not None:
            self.angles = _angles
        else:
            self.angles: dict[str, tuple[float, float]] = {}
        self.has_angles = len(self.angles) > 0

        # Check input types
        if isinstance(q1, Double):
            input = np.array([q1, q2, q3, q4, q5, q6], dtype=np.float64)
            assert len(input.shape) == 1
            assert input.size == 6
        elif isinstance(q1, np.ndarray) and isinstance(q1[0], Double):
            try:
                input = np.array([q1, q2, q3, q4, q5, q6], dtype=np.float64)
                assert len(input.shape) == 2
                assert input.size / 6 == input.shape[1]
            except ValueError or AssertionError:
                raise ValueError(
                    "Components of state vector cannot have different sizes"
                )
        else:
            raise TypeError(
                "Components of state vector must be floating point numbers"
                " or numpy arrays"
            )

        # Initialize frame and state components
        self.frame = frame
        self.q1: Any = np.array([q1], dtype=np.float64).ravel()
        self.q2: Any = np.array([q2], dtype=np.float64).ravel()
        self.q3: Any = np.array([q3], dtype=np.float64).ravel()
        self.q4: Any = np.array([q4], dtype=np.float64).ravel()
        self.q5: Any = np.array([q5], dtype=np.float64).ravel()
        self.q6: Any = np.array([q6], dtype=np.float64).ravel()

        # Convert angles to radians and wrap them if necessary
        if not self.has_angles:
            return None
        if deg:
            for angle in self.angles:
                self.__dict__[angle] = np.deg2rad(self.__dict__[angle])
        if wrap:
            for angle in self.angles:
                low, high = self.angles[angle]
                # self.__dict__[angle] = self._wrap(self.__dict__[angle], low, high)
                self.__dict__[angle] %= high

        return None

    @property
    def size(self) -> int:
        return self.q1.size

    @property
    def scalar(self) -> bool:
        return self.q1.size == 1

    def __getattr__(self, name: str) -> U:

        if name in self.components:
            qi = self.__dict__[self.components[name]]
            return qi[0] if self.scalar else qi

        if name[-4:] == "_deg" and name[:-4] in self.angles:
            qi = np.rad2deg(self.__dict__[self.components[name[:-4]]])
            return qi[0] if self.scalar else qi

        raise AttributeError(
            f"{name} is not a property of {self.__class__.__name__}"
            f" in {self.frame.__name__}"
        )

    def __getitem__(self, index: int | slice) -> Self:
        return type(self)(
            self.q1[index],
            self.q2[index],
            self.q3[index],
            self.q4[index],
            self.q5[index],
            self.q6[index],
        )

    def __iter__(self) -> Iterator[Self]:
        return iter([self.__getitem__(idx) for idx in range(self.size)])

    def __repr__(self) -> str:

        out = f"{self.__class__.__name__} in {self.frame.__name__}\n"
        out += "-" * len(out) + "\n"
        for component in self.components:
            out += f"{component}: {getattr(self, component)}\n"
        return out

    def __add__(self, other: Self | Double | int) -> Self:
        if isinstance(other, Double) or isinstance(other, int):
            other = float(other)
            out = type(self)(
                self.q1 + other,
                self.q2 + other,
                self.q3 + other,
                self.q4 + other,
                self.q5 + other,
                self.q6 + other,
            )
        elif isinstance(other, self.__class__):
            out = type(self)(
                self.q1 + other.q1,
                self.q2 + other.q2,
                self.q3 + other.q3,
                self.q4 + other.q4,
                self.q5 + other.q5,
                self.q6 + other.q6,
            )
        else:
            raise TypeError(
                "Addition is only supported for state vectors and scalars",
            )

        return out

    def __sub__(self, other: Self | Double | int) -> Self:
        if isinstance(other, Double) or isinstance(other, int):
            other = float(other)
            out = type(self)(
                self.q1 - other,
                self.q2 - other,
                self.q3 - other,
                self.q4 - other,
                self.q5 - other,
                self.q6 - other,
            )
        elif isinstance(other, self.__class__):
            out = type(self)(
                self.q1 - other.q1,
                self.q2 - other.q2,
                self.q3 - other.q3,
                self.q4 - other.q4,
                self.q5 - other.q5,
                self.q6 - other.q6,
            )
        else:
            raise TypeError(
                "Subtraction is only supported for state vectors and scalars"
            )

        return out

    def __mul__(self, other: Self | Double | int) -> Self:
        if isinstance(other, Double) or isinstance(other, int):
            other = float(other)
            out = type(self)(
                self.q1 * other,
                self.q2 * other,
                self.q3 * other,
                self.q4 * other,
                self.q5 * other,
                self.q6 * other,
            )
        elif isinstance(other, self.__class__):
            # CHECK: Do I really want to support this?
            out = type(self)(
                self.q1 * other.q1,
                self.q2 * other.q2,
                self.q3 * other.q3,
                self.q4 * other.q4,
                self.q5 * other.q5,
                self.q6 * other.q6,
            )
        else:
            raise TypeError(
                "Multiplication is only supported for state vectors and scalars"
            )

        return out

    def __eq__(self, other) -> bool:
        if isinstance(other, Double) or isinstance(other, int):
            other = float(other)
            return (
                self.q1 == other
                and self.q2 == other
                and self.q3 == other
                and self.q4 == other
                and self.q5 == other
                and self.q6 == other
            )
        elif isinstance(other, self.__class__):
            return (
                np.allclose(self.q1, other.q1, rtol=1e-17)
                and np.allclose(self.q2, other.q2, rtol=1e-17)
                and np.allclose(self.q3, other.q3, rtol=1e-17)
                and np.allclose(self.q4, other.q4, rtol=1e-17)
                and np.allclose(self.q5, other.q5, rtol=1e-17)
                and np.allclose(self.q6, other.q6, rtol=1e-17)
            )
        else:
            raise TypeError(
                "Can only check for equality against state vectors and scalars"
            )

    def append(self, other: Self | Double | int) -> Self:
        if isinstance(other, Double) or isinstance(other, int):
            other = float(other)
            return type(self)(
                np.append(self.q1, other),
                np.append(self.q2, other),
                np.append(self.q3, other),
                np.append(self.q4, other),
                np.append(self.q5, other),
                np.append(self.q6, other),
            )
        elif isinstance(other, self.__class__):
            return type(self)(
                np.append(self.q1, other.q1),
                np.append(self.q2, other.q2),
                np.append(self.q3, other.q3),
                np.append(self.q4, other.q4),
                np.append(self.q5, other.q5),
                np.append(self.q6, other.q6),
            )
        else:
            raise TypeError("Append is only supported for state vectors and scalars")

    def asarray(self) -> Vector:
        return np.array(
            [self.q1, self.q2, self.q3, self.q4, self.q5, self.q6], dtype=np.float64
        )

    def split(self, idx_list: list[int]) -> list[Self]:
        """Split time series of state vectors into intervals

        :param idx_list: List of indices to split the state vector
        """

        print("WARNING: Splitting state vectors is not properly tested\n")
        low = 0
        out = []
        for high in idx_list:
            out.append(self[low:high])
            low = high
        return out

    def save(self, path: str | Path) -> None:
        """Save state vector to Numpy binary file

        :param path: Path to generated file
        """
        np.save(path, self.asarray())
        return None

    @classmethod
    def load(cls, path: str | Path):
        """Load state vector from Numpy binary file

        :param path: Path to file to load
        """
        data = np.load(path)
        return cls(*data)

    @staticmethod
    def _wrap(angle: Double, low: Double = 0.0, high: Double = 2.0 * np.pi) -> Double:
        """Wrap angle to a given interval

        :param angle: Angle to wrap
        :param low: Lower limit of the interval
        :param high: Upper limit of the interval
        """
        # TODO: This should work, but test it
        angle -= low
        angle = angle % (high - low)
        angle += low
        return angle

    def transform(self, frame: type[Frame]) -> Self:
        """Transform state to another reference frame"""
        q1, q2, q3 = self.frame.transform(self.q1, self.q2, self.q3, frame)
        q4, q5, q6 = self.frame.transform(self.q4, self.q5, self.q6, frame)
        out: Any = type(self)(q1, q2, q3, q4, q5, q6, frame=frame)
        return out


class CartesianState[U: (Double, Vector)](GenericState):
    """Cartesian state vector

    Components might be floating point numbers (single state) or numpy arrays
    of them (time series of states).

    Parameters
    -----------
    x, y, z: Double | Vector
        Components of the position vector [m]
    dx, dy, dz: Double | Vector
        Components of the velocity vector [m/s]
    frame: Frame
        Reference frame in which the state vector is defined
    """

    def __init__(
        self, x: U, y: U, z: U, dx: U, dy: U, dz: U, frame: type[Frame] = ICRF
    ) -> None:

        _components = {
            "x": "q1",
            "y": "q2",
            "z": "q3",
            "dx": "q4",
            "dy": "q5",
            "dz": "q6",
        }
        super().__init__(x, y, z, dx, dy, dz, frame=frame, _components=_components)
        return None

    @property
    def r_vec(self) -> Vector:
        return np.array([self.x, self.y, self.z])

    @property
    def v_vec(self) -> Vector:
        return np.array([self.dx, self.dy, self.dz])

    @property
    def r_mag(self) -> U:
        return np.linalg.norm(self.r_vec, axis=0)

    @property
    def v_mag(self) -> U:
        return np.linalg.norm(self.v_vec, axis=0)

    def to_keplerian(self, mu: Double) -> "KeplerianState":
        """Conversion to keplerian state vector

        Algorithm: Practical astrodynamics slides

        :param mu: Gravitational parameter of central body [m^3/s^2]
        """
        # TODO: Ensure that loss of accuracy in np.sum is not relevant
        # TODO: Only allow conversion if frame is valid

        # Semi-major axis
        a = 1.0 / ((2.0 / self.r_mag) - (self.v_mag * self.v_mag / mu))

        # Angular momentum
        h_vec = np.cross(self.r_vec, self.v_vec, axis=0)
        h = np.linalg.norm(h_vec, axis=0)

        # Eccentricity
        e_vec = (np.cross(self.v_vec, h_vec, axis=0) / mu) - (self.r_vec / self.r_mag)
        e = np.linalg.norm(e_vec, axis=0)
        e_uvec = e_vec / e

        # Inclination
        inc = np.arccos(h_vec[2] / h)

        # N vector
        _z_vec = np.array(
            [np.zeros_like(self.x), np.zeros_like(self.y), np.ones_like(self.z)]
        )
        N_vec = np.cross(_z_vec, h_vec, axis=0)
        Nxy = np.sqrt(N_vec[0] * N_vec[0] + N_vec[1] * N_vec[1])
        N_uvec = N_vec / Nxy

        # RAAN
        raan = np.arctan2(N_vec[1] / Nxy, N_vec[0] / Nxy)

        # Argument of periapsis
        sign_aop_condition = np.sum(np.cross(N_uvec, e_vec, axis=0) * h_vec, axis=0) > 0
        sign_aop = 2 * sign_aop_condition - 1
        aop = sign_aop * np.arccos(np.sum(e_uvec * N_uvec, axis=0))

        # True anomaly
        sign_ta_condition = (
            np.sum(np.cross(e_vec, self.r_vec, axis=0) * h_vec, axis=0) > 0
        )
        sign_ta = 2 * sign_ta_condition - 1
        ta = sign_ta * np.arccos(np.sum(self.r_vec * e_uvec / self.r_mag, axis=0))

        return KeplerianState(a, e, inc, raan, aop, ta, deg=False)

    # def transform(self, frame: type[Frame]) -> Self:
    #     """Transform cartesian state to another reference frame"""
    #     x, y, z = self.frame.transform(self.x, self.y, self.z, frame)
    #     dx, dy, dz = self.frame.transform(self.dx, self.dy, self.dz, frame)
    #     out: Any = CartesianState(x, y, z, dx, dy, dz, frame=frame)
    #     return out

    def to_spherical(self) -> "SphericalState":
        """Conversion from cartesian to spherical state"""
        rho, theta, phi = self.frame.to_spherical(self.q1, self.q2, self.q3)
        drho, dtheta, dphi = self.frame.to_spherical(self.q4, self.q5, self.q6)
        out: Any = SphericalState(rho, theta, phi, drho, dtheta, dphi, frame=self.frame)
        return out


class CartesianStateDerivative[U: (Double, Vector)](GenericState):
    """Derivative of cartesian state

    Derivative of a single state of a time series of states of a body in a
    three-dimensional space in terms of cartesian coordinates. Each component
    can be a floating point number (single state) or a numpy array of them
    (time series of states).

    :param dx: X component of the velocity vector [m/s]
    :param dy: Y component of the velocity vector [m/s]
    :param dz: Z component of the velocity vector [m/s]
    :param ddx: X component of the acceleration vector [m/s^2]
    :param ddy: Y component of the acceleration vector [m/s^2]
    :param ddz: Z component of the acceleration vector [m/s^2]
    """

    def __init__(self, dx: U, dy: U, dz: U, ddx: U, ddy: U, ddz: U) -> None:
        super().__init__(dx, dy, dz, ddx, ddy, ddz)
        return None

    @property
    def dx(self) -> U:
        return self.q1[0] if self.scalar else self.q1

    @property
    def dy(self) -> U:
        return self.q2[0] if self.scalar else self.q2

    @property
    def dz(self) -> U:
        return self.q3[0] if self.scalar else self.q3

    @property
    def ddx(self) -> U:
        return self.q4[0] if self.scalar else self.q4

    @property
    def ddy(self) -> U:
        return self.q5[0] if self.scalar else self.q5

    @property
    def ddz(self) -> U:
        return self.q6[0] if self.scalar else self.q6

    @property
    def v_vec(self) -> Vector:
        return np.array([self.dx, self.dy, self.dz])

    @property
    def a_vec(self) -> Vector:
        return np.array([self.ddx, self.ddy, self.ddz])

    @property
    def v_mag(self) -> U:
        return np.linalg.norm(self.v_vec, axis=0)

    @property
    def dv_mag(self) -> U:
        return np.linalg.norm(self.v_vec, axis=0)

    def times_dt(self, dt: Double) -> CartesianState:
        """Change in cartesian state over interval dt

        Returns the change of a cartesian state with this object as derivative
        over a time interval dt. The main use case are numerical integrators,
        in which the state is updated by multiplying its derivative by the
        time step.

        :param dt: Time step
        :return: Change in cartesian elements over interval dt
        """

        return CartesianState(
            self.dx * dt,
            self.dy * dt,
            self.dz * dt,
            self.ddx * dt,
            self.ddy * dt,
            self.ddz * dt,
        )

    def add_acceleration(self, other: Self) -> None:
        """Update the acceleration of the cartesian state derivative

        Update the acceleration of this object by adding the components of the
        acceleration vector of other cartesian state derivative. The use case is
        to add up all the accelerations acting on a body during orbit propagation.
        Since each acceleration is defined by a cartesian state derivative,
        adding them up directly would also update the velocity, which is incorrect.

        :param other: Cartesian state derivative with acceleration to add
        """

        # q4, q5 and q6 are the components of a general state vector that
        # represent the acceleration in CartesianStateDerivative
        self.q4 += other.q4
        self.q5 += other.q5
        self.q6 += other.q6

    def __repr__(self) -> str:
        return (
            "Cartesian state derivative\n"
            "---------------------------------\n"
            f"dx: {self.dx}\n"
            f"dy: {self.dy}\n"
            f"dz: {self.dz}\n"
            f"ddx: {self.ddx}\n"
            f"ddy: {self.ddy}\n"
            f"ddz: {self.ddz}"
        )


class KeplerianState[U: (Double, Vector)](GenericState):
    """Keplerian state vector

    Single state of a time series of states of a body in a three-dimensional
    space in terms of classical keplerian elements. Each component of the state
    vector can be a floating point number (single state) or a numpy array of them
    (time series of states).

    Angles might be given in degrees or radians, but are always converted to
    radians during initialization. For each angle, the method KeplerianState.angle
    will return the value in radians, while KeplerianState.angle_deg will return
    it in degrees.

    Angles can be wrapped if specified by the wrap parameter, which is activated
    by default. Wrapping will limit inclinations to the range [0, pi) and the
    rest of the angles to [0, 2pi).

    :param a: Semi-major axis [m]
    :param e: Eccentricity
    :param i: Inclination [rad]
    :param raan: Right ascension of the ascending node [rad]
    :param aop: Argument of periapsis [rad]
    :param ta: True anomaly [rad]
    :param deg: Angles are given in degrees (True) or radians (False)
    :param wrap: Wrap angles (True) or not (False)
    """

    def __init__(
        self,
        a: U,
        e: U,
        i: U,
        raan: U,
        aop: U,
        ta: U,
        deg: bool = False,
        wrap: bool = True,
        frame: type[Frame] = ICRF,
    ) -> None:
        components = {
            "a": "q1",
            "e": "q2",
            "i": "q3",
            "raan": "q4",
            "aop": "q5",
            "ta": "q6",
        }
        angles = {
            "q3": (0.0, np.pi),
            "q4": (0.0, 2.0 * np.pi),
            "q5": (0.0, 2.0 * np.pi),
            "q6": (0.0, 2.0 * np.pi),
        }
        super().__init__(
            a,
            e,
            i,
            raan,
            aop,
            ta,
            deg=deg,
            wrap=wrap,
            frame=frame,
            _components=components,
            _angles=angles,
        )

        return None

    @property
    def E(self) -> U:
        """Eccentric anomaly"""
        out: Any = 2.0 * np.arctan2(
            np.sqrt(1.0 - self.e) * np.sin(0.5 * self.ta),
            np.sqrt(1.0 + self.e) * np.cos(0.5 * self.ta),
        )
        return out

    @property
    def M(self) -> U:
        """Mean anomaly"""
        out: Any = self.E - self.e * np.sin(self.E)
        return out

    def period(self, mu: Double) -> U:
        """Orbital period"""
        out: Any = 2.0 * np.pi * np.sqrt(self.a * self.a * self.a / mu)
        return out

    def __sub__(self, other: object) -> Self:
        print(
            "WARNING: Keplerian state subtraction is not properly tested\n"
            "It might produce unexpected results"
        )

        if not isinstance(other, KeplerianState):
            raise TypeError("Subtraction is only supported for keplerian state vectors")

        di = np.unwrap(self.q3, period=np.pi) - np.unwrap(other.q3, period=np.pi)
        draan = np.unwrap(self.q4, period=2.0 * np.pi) - np.unwrap(
            other.q4, period=2.0 * np.pi
        )
        daop = np.unwrap(self.q5, period=2.0 * np.pi) - np.unwrap(
            other.q5, period=2.0 * np.pi
        )
        dta = np.unwrap(self.q6, period=2.0 * np.pi) - np.unwrap(
            other.q6, period=2.0 * np.pi
        )

        return type(self)(
            self.q1 - other.q1,
            self.q2 - other.q2,
            di,
            draan,
            daop,
            dta,
            frame=self.frame,
        )

    def to_cartesian(self, mu: Double) -> "CartesianState":
        """Conversion to cartesian state vector

        Algorithm: Practical astrodynamics slides

        :param mu: Gravitational parameter of central body [m^3/s^2]
        """

        # Auxiliary trigonometric relations
        cos_Omega = np.cos(self.raan)
        sin_Omega = np.sin(self.raan)
        cos_omega = np.cos(self.aop)
        sin_omega = np.sin(self.aop)
        cos_i = np.cos(self.i)
        sin_i = np.sin(self.i)
        cos_theta = np.cos(self.ta)
        sin_theta = np.sin(self.ta)

        l1 = cos_Omega * cos_omega - sin_Omega * sin_omega * cos_i
        l2 = -cos_Omega * sin_omega - sin_Omega * cos_omega * cos_i
        m1 = sin_Omega * cos_omega + cos_Omega * sin_omega * cos_i
        m2 = -sin_Omega * sin_omega + cos_Omega * cos_omega * cos_i
        n1 = sin_omega * sin_i
        n2 = cos_omega * sin_i

        # Orbital radius
        r = self.a * (1.0 - self.e * self.e) / (1.0 + self.e * cos_theta)

        # Position in orbital plane
        xi = r * cos_theta
        eta = r * sin_theta

        # Position in 3D space
        x = l1 * xi + l2 * eta
        y = m1 * xi + m2 * eta
        z = n1 * xi + n2 * eta

        # Angular momentum
        H = np.sqrt(mu * self.a * (1.0 - self.e * self.e))

        # Velocity
        common = mu / H
        e_cos_theta = self.e + cos_theta

        dx = common * (l2 * e_cos_theta - l1 * sin_theta)
        dy = common * (m2 * e_cos_theta - m1 * sin_theta)
        dz = common * (n2 * e_cos_theta - n1 * sin_theta)

        return CartesianState(x, y, z, dx, dy, dz, frame=self.frame)


class KeplerianStateDerivative[U: (Double, Vector)](GenericState):
    """Derivative of keplerian state

    Derivative of a single state of a time series of states of a body in a
    three-dimensional space in terms of classical keplerian elements. Each
    component can be a floating point number (single state) or a numpy array
    of them (time series of states).

    Angular velocities might be given in degrees or radians per second, but are
    always converted to radians per second during initialization. For each angular
    velocity, the method KeplerianStateDerivative.angvel will return the value in
    radians per second, while KeplerianStateDerivative.angvel_deg will return it
    in degrees per second.

    :param da: Rate of change of semi-major axis [m/s]
    :param de: Rate of change of eccentricity [1/s]
    :param di: Rate of change of inclination [rad/s]
    :param draan: Rate of change of right ascension of the ascending node [rad/s]
    :param daop: Rate of change of argument of periapsis [rad/s]
    :param dta: Rate of change of true anomaly [rad/s]
    :param deg: Angular velocities are given in degrees per second (True) or
        radians  per second (False)
    """

    def __init__(
        self,
        da: U,
        de: U,
        di: U,
        draan: U,
        daop: U,
        dta: U,
        deg: bool = True,
    ) -> None:
        super().__init__(da, de, di, draan, daop, dta)

        if not deg:
            self.q3 = np.rad2deg(self.q3)
            self.q4 = np.rad2deg(self.q4)
            self.q5 = np.rad2deg(self.q5)
            self.q6 = np.rad2deg(self.q6)

        return None

    @property
    def da(self) -> U:
        return self.q1[0] if self.scalar else self.q1

    @property
    def de(self) -> U:
        return self.q2[0] if self.scalar else self.q2

    @property
    def di(self) -> U:
        return self.q3[0] if self.scalar else self.q3

    @property
    def draan(self) -> U:
        return self.q4[0] if self.scalar else self.q4

    @property
    def daop(self) -> U:
        return self.q5[0] if self.scalar else self.q5

    @property
    def dta(self) -> U:
        return self.q6[0] if self.scalar else self.q6

    def times_dt(self, dt: Double) -> KeplerianState:
        """Change in keplerian state over interval dt

        Returns the change of a keplerian state with this object as derivative
        over a time interval dt. The main use case are numerical integrators,
        in which the state is updated by multiplying its derivative by the
        time step.

        :param dt: Time step
        :return: Change in keplerian elements over interval dt
        """

        return KeplerianState(
            self.da * dt,
            self.de * dt,
            self.di * dt,
            self.draan * dt,
            self.daop * dt,
            self.dta * dt,
        )

    def __repr__(self) -> str:
        return (
            "Keplerian state derivative\n"
            "---------------------------------\n"
            f"da: {self.da}\n"
            f"de: {self.de}\n"
            f"di: {self.di}\n"
            f"dRAAN: {self.draan}\n"
            f"dAoP: {self.daop}\n"
            f"dTA: {self.dta}"
        )


class SphericalState[U: (Double, Vector)](GenericState):

    def __init__(
        self,
        rho: U,
        theta: U,
        phi: U,
        drho: U,
        dtheta: U,
        dphi: U,
        deg: bool = False,
        wrap: bool = True,
        frame: type[Frame] = ICRF,
    ) -> None:

        components = frame.variables
        angles = frame.angles

        super().__init__(
            rho,
            theta,
            phi,
            drho,
            dtheta,
            dphi,
            deg=deg,
            wrap=wrap,
            frame=frame,
            _components=components,
            _angles=angles,
        )

        return None

    def __add__(self, other: object) -> Self:
        raise NotImplementedError(
            "Addition is still not supported for spherical state vectors"
        )

    def __sub__(self, other: object) -> Self:
        raise NotImplementedError(
            "Subtraction is still not supported for spherical state vectors"
        )

    def transform(self, frame: type[Frame]) -> Self:
        """Transform spherical state to another reference frame"""
        cstate = self.to_cartesian()
        q1, q2, q3 = self.frame.transform(cstate.q1, cstate.q2, cstate.q3, frame)
        q4, q5, q6 = self.frame.transform(cstate.q4, cstate.q5, cstate.q6, frame)
        out: Any = CartesianState(q1, q2, q3, q4, q5, q6, frame=frame).to_spherical()
        return out

    def to_cartesian(self) -> "CartesianState":
        """Conversion from spherical to cartesian state"""
        x, y, z = self.frame.to_cartesian(self.q1, self.q2, self.q3)
        dx, dy, dz = self.frame.to_cartesian(self.q4, self.q5, self.q6)
        out: Any = CartesianState(x, y, z, dx, dy, dz, frame=self.frame)
        return out

    def to_keplerian(self, mu: Double) -> "KeplerianState":
        """Conversion from spherical to keplerian state"""
        return self.to_cartesian().to_keplerian(mu)


class CartesianPosition[U: (Double, Vector)](GenericState):
    """Three dimensional position vector in cartesian coordinates

    Components might be floating point numbers (single vector) or numpy arrays
    of them (time series of vectors).

    :param x: X coordinate [m]
    :param y: Y coordinate [m]
    :param z: Z coordinate [m]
    """

    def __init__(self, x: U, y: U, z: U, *_, frame: type[Frame] = ICRF) -> None:

        components = {"x": "q1", "y": "q2", "z": "q3"}
        null = 0.0 * x
        super().__init__(x, y, z, null, null, null, frame=frame, _components=components)

        return None

    def asarray(self) -> Vector:
        return np.array([self.x, self.y, self.z])

    @property
    def mag(self) -> U:
        return np.linalg.norm(self.asarray(), axis=0)

    def to_spherical(self) -> "SphericalPosition":
        rho, theta, phi = self.frame.to_spherical(self.x, self.y, self.z)
        out: Any = SphericalPosition(rho, theta, phi, frame=self.frame)
        return out


class CartesianVelocity[U: (Double, Vector)](GenericState):
    """Three dimensional velocity vector in cartesian coordinates

    Components might be floating point numbers (single vector) or numpy arrays
    of them (time series of vectors).

    :param dx: X component of the velocity vector [m/s]
    :param dy: Y component of the velocity vector [m/s]
    :param dz: Z component of the velocity vector [m/s]
    """

    def __init__(self, dx: U, dy: U, dz: U, *_, frame: type[Frame] = ICRF) -> None:

        components = {"dx": "q1", "dy": "q2", "dz": "q3"}
        null = 0.0 * dx
        super().__init__(
            dx, dy, dz, null, null, null, frame=frame, _components=components
        )
        return None

    def asarray(self) -> Vector:
        return np.array([self.dx, self.dy, self.dz])

    @property
    def mag(self) -> U:
        return np.linalg.norm(self.asarray(), axis=0)


class SphericalPosition[U: (Double, Vector)](GenericState):
    """Three dimensional position vector in spherical coordinates

    Components might be floating point numbers (single vector) or numpy arrays
    of them (time series of vectors).

    Angles can be passed in radians (default) or degrees (deg=True), but they
    are always converted to radians during initialization. For each angle, the
    method SphericalPosition.angle will return the value in radians, while
    SphericalPosition.angle_deg will return it in degrees.

    :param range: Radial distance [m]
    :param theta: Horizontal angular distance [rad / deg]
    :param phi: Vertical angular distance [rad / deg]
    :param deg: Angles are given in degrees (True) or radians (False)
    :param wrap: Wrap angles (True) or not (False)
    """

    def __init__(
        self,
        rho: U,
        theta: U,
        phi: U,
        deg: bool = False,
        wrap: bool = True,
        frame: type[Frame] = ICRF,
    ) -> None:

        _keys = list(frame.variables.keys())
        components = {_keys[0]: "q1", _keys[1]: "q2", _keys[2]: "q3"}

        null = 0.0 * rho
        super().__init__(
            rho,
            theta,
            phi,
            null,
            null,
            null,
            frame=frame,
            deg=deg,
            wrap=wrap,
            _components=components,
            _angles=frame.angles,
        )

        return None

    def asarray(self) -> Vector:
        return np.array([self.__getattr__(var) for var in self.frame.variables])

    def to_cartesian(self) -> "CartesianPosition":
        x, y, z = self.frame.to_cartesian(self.q1, self.q2, self.q3)
        out: Any = CartesianPosition(x, y, z, frame=self.frame)
        return out

    def transform(self, frame: type[Frame]) -> Self:
        """Transform spherical position to another reference frame"""
        out: Any = self.to_cartesian().transform(frame).to_spherical()
        return out
